from __future__ import annotations

from typing import TYPE_CHECKING
import logging
import weakref
from app_model import Action
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt
from superqt import QIconifyIcon

from himena.utils.misc import lru_cache

if TYPE_CHECKING:
    from himena.widgets import MainWindow

_LOGGER = logging.getLogger(__name__)


class QCommandHistory(QtW.QWidget):
    """List of command history.

    The executed commands are listed in the widget, and can be re-executed by clicking
    the button.
    """

    def __init__(self, ui: MainWindow):
        super().__init__()
        layout = QtW.QVBoxLayout(self)
        self._command_list = QCommandList(ui)
        layout.addWidget(self._command_list)
        ui.model_app.commands.executed.connect(self._command_executed)
        self._ui_ref = weakref.ref(ui)

    def _command_executed(self, command_id: str) -> None:
        _LOGGER.info("Command executed: %s", command_id)
        num = len(self._ui_ref()._history_command) - 1
        self._command_list.model().beginInsertRows(QtCore.QModelIndex(), num, num)
        self._command_list.model().insertRow(num, QtCore.QModelIndex())
        self._command_list._update_index_widgets()
        self._command_list.model().endInsertRows()

    def deleteLater(self):
        self._command_list.model().beginRemoveRows(
            QtCore.QModelIndex(), 0, self._command_list.model().rowCount() - 1
        )
        self._ui_ref = lambda: None
        self._command_list.model().removeRows(0, self._command_list.model().rowCount())
        self._command_list.model().endRemoveRows()
        return super().deleteLater()


class QCommandList(QtW.QListView):
    current_index_changed = QtCore.Signal(int)

    def __init__(self, ui: MainWindow, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        model = QCommandListModel(ui)
        self.setModel(model)
        self.setEditTriggers(QtW.QAbstractItemView.EditTrigger.NoEditTriggers)
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.SingleSelection)

    def model(self) -> QCommandListModel:
        return super().model()

    def currentChanged(
        self, current: QtCore.QModelIndex, previous: QtCore.QModelIndex
    ) -> None:
        super().currentChanged(current, previous)
        row = current.row()
        if row >= 0:
            self.current_index_changed.emit(row)

    def _update_index_widgets(self):
        for row in range(self.model().rowCount()):
            index = self.model().index(row)
            if not index.isValid():
                return None
            if action := self.model()._action_at(row):
                id = action.id
                title = action.title
            else:
                id = ""
                title = ""
            if widget := self.indexWidget(index):
                widget.setText(title)
            else:
                widget = QCommandIndexWidget(id, title, self)
                self.setIndexWidget(index, widget)
                widget.btn_clicked.connect(self._execute_action_at_widget)

    def _find_index_widget(
        self, widget: QCommandIndexWidget
    ) -> QtCore.QModelIndex | None:
        for row in range(self.model().rowCount()):
            index = self.model().index(row)
            if self.indexWidget(index) == widget:
                if not index.isValid():
                    return None
                return index

    def _execute_action_at_widget(self, widget: QCommandIndexWidget):
        index = self._find_index_widget(widget)
        if index is None:
            return
        if action := self.model()._action_at(index.row()):
            if ui := self.model()._ui_ref():
                ui.exec_action(action.id)

    if TYPE_CHECKING:

        def indexWidget(
            self, index: QtCore.QModelIndex
        ) -> QCommandIndexWidget | None: ...


def _color(light_background: bool):
    if light_background:
        color = "#222222"
    else:
        color = "#E6E6E6"
    return color


@lru_cache(maxsize=1)
def _icon_run(light_background: bool) -> QIconifyIcon:
    return QIconifyIcon("fa:play", color=_color(light_background))


@lru_cache(maxsize=1)
def _icon_copy(light_background: bool) -> QIconifyIcon:
    return QIconifyIcon(
        "heroicons-outline:clipboard-copy", color=_color(light_background)
    )


def make_btn(tooltip: str):
    btn = QtW.QToolButton()
    btn.setObjectName("QCommandHistory-RunButton")
    btn.setIcon(QtGui.QIcon())
    btn.setFixedWidth(20)
    btn.setToolTip(tooltip)
    return btn


class QCommandIndexWidget(QtW.QWidget):
    btn_clicked = QtCore.Signal(object)

    def __init__(self, id: str, text: str, listwidget: QCommandList):
        super().__init__()
        self._action_id = id
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(5, 0, 0, 0)
        layout.setSpacing(0)
        self._label = QtW.QLabel(text)
        self._btn_run = make_btn("Run this command")
        self._btn_copy = make_btn("Copy this command ID")
        layout.addWidget(self._label, stretch=100, alignment=Qt.AlignmentFlag.AlignLeft)
        layout.addWidget(self._btn_copy, alignment=Qt.AlignmentFlag.AlignRight)
        layout.addWidget(self._btn_run, alignment=Qt.AlignmentFlag.AlignRight)
        self.setMouseTracking(True)
        self._listwidget_ref = weakref.ref(listwidget)
        self._btn_run.clicked.connect(self._emit_run_clicked)
        self._btn_copy.clicked.connect(self._emit_copy_clicked)

    def setText(self, text: str):
        self._label.setText(text)

    def _emit_run_clicked(self):
        self.btn_clicked.emit(self)

    def _emit_copy_clicked(self):
        if self._action_id and (clipboard := QtW.QApplication.clipboard()):
            clipboard.setText(self._action_id)

    def _is_light_background(self) -> bool:
        try:
            return self._listwidget_ref().model()._ui_ref().theme.is_light_background()
        except Exception:
            return False

    def set_button_visible(self, visible: bool):
        self._btn_run.setEnabled(visible)
        if visible:
            light_bg = self._is_light_background()
            self._btn_run.setIcon(QtGui.QIcon(_icon_run(light_bg)))
            self._btn_copy.setIcon(QtGui.QIcon(_icon_copy(light_bg)))
            self._btn_run.setCursor(Qt.CursorShape.PointingHandCursor)
            self._btn_copy.setCursor(Qt.CursorShape.PointingHandCursor)
        else:
            self._btn_run.setIcon(QtGui.QIcon())
            self._btn_copy.setIcon(QtGui.QIcon())
            self._btn_run.setCursor(Qt.CursorShape.ArrowCursor)
            self._btn_copy.setCursor(Qt.CursorShape.ArrowCursor)

    def enterEvent(self, event: QtCore.QEvent) -> None:
        self._enter_event()
        return super().enterEvent(event)

    def _enter_event(self):
        listwidget = self._listwidget_ref()
        if listwidget is None:
            return
        ui = listwidget.model()._ui_ref()
        index = listwidget._find_index_widget(self)
        action = listwidget.model()._action_at(index.row())
        if ui is None or action is None:
            return
        # check enablement
        if action.enablement is None:
            enabled = True
        else:
            ctx = ui._ctx_keys.dict()
            enabled = action.enablement.eval(ctx)
        self.set_button_visible(enabled)
        return

    def leaveEvent(self, event: QtCore.QEvent) -> None:
        self.set_button_visible(False)
        return super().leaveEvent(event)

    def _make_drag(self) -> QtGui.QDrag:
        qdrag = QtGui.QDrag(self)

        # prepare mime data
        mime = QtCore.QMimeData()
        mime.setText(self._action_id)
        mime.setData("text/plain", self._action_id.encode())
        mime.setData("text/command-id", self._action_id.encode())
        qdrag.setMimeData(mime)

        # prepare pixmap
        pixmap = QtGui.QPixmap(self.size())
        self.render(pixmap)
        qdrag.setPixmap(pixmap)
        return qdrag

    def mouseMoveEvent(self, a0):
        super().mouseMoveEvent(a0)
        if a0.buttons() == Qt.MouseButton.LeftButton:
            qdrag = self._make_drag()
            qdrag.exec(Qt.DropAction.MoveAction)


class QCommandListModel(QtCore.QAbstractListModel):
    def __init__(self, ui: MainWindow, parent=None):
        super().__init__(parent)
        self._ui_ref = weakref.ref(ui)

    def rowCount(self, parent=None):
        if ui := self._ui_ref():
            return ui._history_command.len()
        return 0

    def _action_at(self, row: int) -> Action | None:
        """app-model Action at the given row."""
        if ui := self._ui_ref():
            command_id = ui._history_command.get(row)
            if command_id is None:
                return None
            return ui.model_app.registered_actions.get(command_id)
        return None

    def data(self, index, role):
        if not index.isValid():
            return QtCore.QVariant()
        if role == Qt.ItemDataRole.ToolTipRole:
            if action := self._action_at(index.row()):
                return action.tooltip
        elif role == Qt.ItemDataRole.StatusTipRole:
            if action := self._action_at(index.row()):
                return action.status_tip
        elif role == Qt.ItemDataRole.DisplayRole:
            return ""
        return QtCore.QVariant()



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtCore

from himena_builtins.qt.widgets._table_components._selection_model import Index
from himena_builtins.qt.widgets.table import QSpreadsheet
from himena_builtins.qt.widgets.dict import QDictOfWidgetEdit
from himena_builtins.qt.widgets._table_components import QSelectionRangeEdit
from himena import MainWindow
from himena.types import WidgetDataModel
from himena.consts import StandardType
from himena.plugins import validate_protocol


class QExcelEdit(QDictOfWidgetEdit):
    """Built-in Excel File Editor.

    ## Basic Usage

    This widget is used to view and edit Excel books (stack of spreadsheets). It works
    almost like a tabbed list of built-in spreadsheet for simple table types. Note that
    this widget is not designed for full replacement of Excel software. Rich text,
    formulas, and other advanced features are not supported.

    ## Drag and Drop

    Dragging a tab will provide a model of type `StandardType.TABLE` ("table").
    `Ctrl + left_button` or `middle button` are assigned to the drag event.
    """

    __himena_widget_id__ = "builtins:QExcelEdit"
    __himena_display_name__ = "Built-in Excel File Editor"

    def __init__(self, ui: MainWindow):
        super().__init__()
        self._ui = ui
        self._model_type_component = StandardType.TABLE
        self._model_type = StandardType.EXCEL
        self._control = QExcelTableStackControl()
        self._extension_default = ".xlsx"

    def _default_widget(self) -> QSpreadsheet:
        table = QSpreadsheet(self._ui)
        table.update_model(WidgetDataModel(value=None, type=StandardType.TABLE))
        table.setHeaderFormat(QSpreadsheet.HeaderFormat.Alphabetic)
        return table

    @validate_protocol
    def control_widget(self) -> QExcelTableStackControl:
        return self._control


_R_CENTER = QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter


class QExcelTableStackControl(QtW.QWidget):
    def __init__(self):
        super().__init__()
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(_R_CENTER)
        # self._header_format = QtW.QComboBox()
        # self._header_format.addItems(["0, 1, 2, ...", "1, 2, 3, ...", "A, B, C, ..."])
        self._value_line_edit = QtW.QLineEdit()
        self._label = QtW.QLabel("")
        self._label.setAlignment(_R_CENTER)
        self._selection_range = QSelectionRangeEdit()
        # layout.addWidget(self._header_format)

        # toolbuttons
        self._insert_menu_button = QtW.QPushButton()
        self._insert_menu_button.setText("Ins")  # or "icons8:plus"
        self._insert_menu_button.setMenu(self._make_insert_menu())
        self._remove_menu_button = QtW.QPushButton()
        self._remove_menu_button.setText("Rem")
        self._remove_menu_button.setMenu(self._make_delete_menu())

        layout.addWidget(self._value_line_edit)
        layout.addWidget(self._insert_menu_button)
        layout.addWidget(self._remove_menu_button)
        layout.addWidget(self._label)
        layout.addWidget(self._selection_range)
        self._value_line_edit.editingFinished.connect(self.update_for_editing)

    def update_for_component(self, table: QSpreadsheet | None):
        if table is None:
            return
        shape = table.model()._arr.shape
        self._label.setText(f"Shape {shape!r}")
        self._selection_range.connect_table(table)
        table._selection_model.moved.connect(self.update_for_current_index)
        self.update_for_current_index(
            table._selection_model.current_index, table._selection_model.current_index
        )
        return None

    @property
    def _current_table(self) -> QSpreadsheet | None:
        return self._selection_range._qtable

    def update_for_current_index(self, old: Index, new: Index):
        qtable = self._current_table
        if qtable is None:
            return None
        qindex = qtable.model().index(new.row, new.column)
        text = qtable.model().data(qindex)
        if not isinstance(text, str):
            text = ""
        self._value_line_edit.setText(text)
        return None

    def update_for_editing(self):
        qtable = self._current_table
        if qtable is None:
            return None
        text = self._value_line_edit.text()
        index = qtable._selection_model.current_index
        qindex = qtable.model().index(index.row, index.column)
        qtable.model().setData(qindex, text, QtCore.Qt.ItemDataRole.EditRole)
        qtable.setFocus()
        return None

    def _make_insert_menu(self):
        menu = QtW.QMenu(self)
        menu.addAction("Row above", self._insert_row_above)
        menu.addAction("Row below", self._insert_row_below)
        menu.addAction("Column left", self._insert_column_left)
        menu.addAction("Column right", self._insert_column_right)
        return menu

    def _make_delete_menu(self):
        menu = QtW.QMenu(self)
        menu.addAction("Rows", self._remove_selected_rows)
        menu.addAction("Columns", self._remove_selected_columns)
        return menu

    def _insert_row_above(self):
        if qtable := self._current_table:
            qtable._insert_row_above()
        return None

    def _insert_row_below(self):
        if qtable := self._current_table:
            qtable._insert_row_below()
        return None

    def _insert_column_left(self):
        if qtable := self._current_table:
            qtable._insert_column_left()
        return None

    def _insert_column_right(self):
        if qtable := self._current_table:
            qtable._insert_column_right()
        return None

    def _remove_selected_rows(self):
        if qtable := self._current_table:
            qtable._remove_selected_rows()
        return None

    def _remove_selected_columns(self):
        if qtable := self._current_table:
            qtable._remove_selected_columns()
        return None



from __future__ import annotations

from typing import TYPE_CHECKING
import numpy as np
from qtpy import QtWidgets as QtW
from qtpy import QtCore, QtGui
from superqt import QLabeledDoubleSlider, QToggleSwitch, QElidingLabel
from superqt.utils import qthrottled

from himena.consts import DefaultFontFamily
from himena.qt._utils import qsignal_blocker, ndarray_to_qimage
from himena_builtins.qt.widgets._image_components import QHistogramView
from himena.utils.enum import StrEnum

if TYPE_CHECKING:
    from himena_builtins.qt.widgets.image import QImageViewBase, ChannelInfo


class ImageType(StrEnum):
    SINGLE = "Single"
    RGB = "RGB"
    MULTI = "Multi"
    OTHERS = "Others"


class ComplexMode(StrEnum):
    REAL = "Real"
    IMAG = "Imag"
    ABS = "Abs"
    LOG_ABS = "Log Abs"
    PHASE = "Phase"


class ChannelMode(StrEnum):
    COMP = "Comp."
    MONO = "Mono"
    GRAY = "Gray"


class RGBMode(StrEnum):
    COLOR = "Color"
    GRAY = "Gray"


class QImageViewControlBase(QtW.QWidget):
    def __init__(self, image_view: QImageViewBase):
        super().__init__()
        self._image_view = image_view
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self._interp_check_box = QInterpolationSwitch(self)
        self._interp_check_box.toggled.connect(self._interpolation_changed)

        self._hover_info = QElidingLabel()
        self._hover_info.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter
        )
        self._hover_info.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )

        for wdt in self._widgets_to_add():
            layout.addWidget(wdt)

    def _widgets_to_add(self) -> list[QtW.QWidget]:
        return [self._hover_info, self._interp_check_box]

    def _interpolation_changed(self, checked: bool):
        self._image_view._img_view.setSmoothing(checked)

    def update_rgb_channel_dtype(self, is_rgb: bool, nchannels: int, dtype):
        pass


class QImageViewControl(QImageViewControlBase):
    def __init__(self, image_view: QImageViewBase):
        self._complex_mode_old = "Abs"
        self._cmp_mode_combo = QtW.QComboBox()
        self._cmp_mode_combo.addItems(
            [ComplexMode.REAL, ComplexMode.IMAG, ComplexMode.ABS, ComplexMode.LOG_ABS,
             ComplexMode.PHASE]
        )  # fmt: skip
        self._cmp_mode_combo.setCurrentIndex(2)
        self._cmp_mode_combo.setToolTip("Method to display complex data")

        self._chn_vis = QChannelToggleSwitches()

        self._chn_mode_combo = QtW.QComboBox()
        self._chn_mode_combo.addItems([""])
        self._chn_mode_combo.setToolTip("Method to display multi-channel data")
        self._image_type = ImageType.OTHERS

        self._auto_cont_btn = QtW.QPushButton("Auto")
        self._auto_cont_btn.setToolTip("Auto contrast")

        self._histogram = QHistogramView()
        self._histogram.setFixedWidth(120)
        super().__init__(image_view)
        self._cmp_mode_combo.hide()
        self._chn_mode_combo.hide()
        self._cmp_mode_combo.currentTextChanged.connect(self._on_complex_mode_change)
        self._chn_vis.stateChanged.connect(self._on_channel_visibility_change)
        self._histogram.clim_changed.connect(self._clim_changed)
        self._auto_cont_btn.clicked.connect(self._auto_contrast)
        self._chn_mode_combo.currentTextChanged.connect(self._on_channel_mode_change)

    def _widgets_to_add(self) -> list[QtW.QWidget]:
        return [
            self._hover_info, self._cmp_mode_combo, self._chn_vis,
            self._chn_mode_combo, self._auto_cont_btn, self._histogram,
            self._interp_check_box,
        ]  # fmt: skip

    def update_rgb_channel_dtype(self, is_rgb: bool, nchannels: int, dtype):
        dtype = np.dtype(dtype)
        if is_rgb:
            kind = ImageType.RGB
        elif nchannels > 1:
            kind = ImageType.MULTI
        else:
            kind = ImageType.SINGLE
        if kind != self._image_type:
            if kind is ImageType.RGB:
                self._chn_mode_combo.clear()
                self._chn_mode_combo.addItems([RGBMode.COLOR, RGBMode.GRAY])
                self._chn_mode_combo.show()
                self._chn_vis.hide()
            elif kind is ImageType.MULTI:
                self._chn_mode_combo.clear()
                self._chn_mode_combo.addItems(
                    [ChannelMode.COMP, ChannelMode.MONO, ChannelMode.GRAY]
                )
                self._chn_mode_combo.show()
                self._chn_vis.show()
            else:
                self._chn_mode_combo.clear()
                self._chn_mode_combo.addItems([""])
                self._chn_mode_combo.hide()
                self._chn_vis.hide()
            self._image_type = kind
            self._chn_mode_combo.setCurrentIndex(0)
        self._cmp_mode_combo.setVisible(dtype.kind == "c")
        if dtype.kind in "uib":
            self._histogram.setValueFormat(".0f")
        else:
            self._histogram.setValueFormat(".3g")
        return None

    def complex_transform(self, arr: np.ndarray) -> np.ndarray:
        """Transform complex array according to the current complex mode."""
        if self._cmp_mode_combo.currentText() == ComplexMode.REAL:
            return arr.real
        if self._cmp_mode_combo.currentText() == ComplexMode.IMAG:
            return arr.imag
        if self._cmp_mode_combo.currentText() == ComplexMode.ABS:
            return np.abs(arr)
        if self._cmp_mode_combo.currentText() == ComplexMode.LOG_ABS:
            return np.log(np.abs(arr) + 1e-6)
        if self._cmp_mode_combo.currentText() == ComplexMode.PHASE:
            return np.angle(arr)
        return arr

    @qthrottled(timeout=100)
    def _clim_changed(self, clim: tuple[float, float]):
        view = self._image_view
        ch = view.current_channel()
        ch.clim = clim
        idx = ch.channel_index or 0
        imtup = view._current_image_slices[idx]
        with qsignal_blocker(self._histogram):
            _grays = (RGBMode.GRAY, ChannelMode.GRAY)
            if imtup.visible:
                arr = ch.transform_image(
                    view._current_image_slices[idx].arr,
                    complex_transform=self.complex_transform,
                    is_rgb=view._is_rgb,
                    is_gray=self._chn_mode_combo.currentText() in _grays,
                )
            else:
                arr = None
            view._img_view.set_array(idx, arr)

    def _on_channel_mode_change(self, mode: str):
        self._chn_vis.setVisible(mode == ChannelMode.COMP)
        self._on_channel_visibility_change()

    def _channel_visibility(self) -> list[bool]:
        caxis = self._image_view._channel_axis
        if caxis is None:
            return [True]  # No channels, always visible
        is_composite = self._chn_mode_combo.currentText() == ChannelMode.COMP
        if is_composite:
            visibilities = self._chn_vis._check_states()
        else:
            visibilities = [False] * len(self._chn_vis._toggle_switches)
            sl = self._image_view._dims_slider.value()
            ith_channel = sl[caxis]
            if len(visibilities) <= ith_channel:
                return [True] * len(sl)  # before initialization
            visibilities[ith_channel] = True
        return visibilities

    def _on_channel_visibility_change(self):
        visibilities = self._channel_visibility()
        self._image_view._update_image_visibility(visibilities)

    def _on_complex_mode_change(self):
        cur = self._cmp_mode_combo.currentText()
        self._image_view._reset_image()
        self._complex_mode_old = cur
        # TODO: auto contrast and update colormap

    def _auto_contrast(self):
        min_ = float("inf")
        max_ = -float("inf")
        for item in self._histogram._hist_items:
            min_ = min(item._edges[0], min_)
            max_ = max(item._edges[-1], max_)
        if np.isinf(min_) or np.isinf(max_):
            return

        view = self._image_view
        sl = view._dims_slider.value()
        img_slice = view._get_image_slice_for_channel(sl)
        if img_slice.dtype.kind == "c":
            img_slice = self.complex_transform(img_slice)
        ch = view.current_channel(sl)
        ch.clim = (min_, max_)
        self._histogram.set_view_range(min_, max_)
        view._set_image_slice(img_slice, ch)
        self._histogram.set_clim((min_, max_))


class QImageLabelViewControl(QImageViewControlBase):
    def __init__(self, image_view: QImageViewBase):
        self._opacity_slider = QLabeledDoubleSlider(QtCore.Qt.Orientation.Horizontal)
        self._opacity_slider.setValue(0.6)
        self._opacity_slider.setFixedWidth(120)
        self._opacity_slider.setRange(0.0, 1.0)
        self._opacity_slider.setToolTip("Opacity of the label colors")
        super().__init__(image_view)
        self._opacity_slider.valueChanged.connect(image_view._reset_image)

    def _widgets_to_add(self):
        return [self._hover_info, self._opacity_slider, self._interp_check_box]


class QInterpolationSwitch(QtW.QAbstractButton):
    """Button for the interpolation mode."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setCheckable(True)
        self.setCursor(QtCore.Qt.CursorShape.PointingHandCursor)
        self.setToolTip("Toggle interpolation mode")

    def paintEvent(self, a0: QtGui.QPaintEvent | None) -> None:
        p = QtGui.QPainter(self)
        p.drawPixmap(0, 0, self._make_pixmap(self.isChecked()))
        p.end()

    def minimumSizeHint(self) -> QtCore.QSize:
        return self.sizeHint()

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(20, 20)

    def _make_pixmap(self, state: bool) -> QtGui.QPixmap:
        img = np.array(
            [[1, 0, 1, 0], [0, 1, 0, 1], [1, 0, 1, 0], [0, 1, 0, 1]],
            dtype=np.uint8,
        )
        img = img * 188 + 34
        qimg = ndarray_to_qimage(img)
        qpixmap = QtGui.QPixmap.fromImage(qimg)
        if state:
            tr = QtCore.Qt.TransformationMode.SmoothTransformation
        else:
            tr = QtCore.Qt.TransformationMode.FastTransformation
        return qpixmap.scaled(
            self.width(), self.height(), QtCore.Qt.AspectRatioMode.KeepAspectRatio, tr
        )


class QChannelToggleSwitch(QToggleSwitch):
    """Toggle switch for channel visibility."""

    def __init__(self, channel: ChannelInfo):
        super().__init__()
        self._channel_info = channel
        self.setChecked(True)

    def set_channel(self, channel: ChannelInfo):
        self._channel_info = channel
        self.setText(channel.name)
        self._channel_on_color = QtGui.QColor.fromRgbF(*channel.colormap(0.5))

    def drawGroove(self, painter, rect, option):
        painter.setPen(QtCore.Qt.PenStyle.NoPen)
        is_checked = option.state & QtW.QStyle.StateFlag.State_On
        painter.setBrush(self._channel_on_color if is_checked else option.off_color)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, True)
        painter.setOpacity(0.65)
        painter.drawRect(rect)

    def drawHandle(self, painter, rect, option):
        painter.setPen(QtCore.Qt.PenStyle.SolidLine)
        is_checked = option.state & QtW.QStyle.StateFlag.State_On
        painter.setBrush(self._channel_on_color if is_checked else option.off_color)
        painter.setOpacity(1.0)
        painter.drawRect(rect)


class QChannelToggleSwitches(QtW.QScrollArea):
    stateChanged = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        central = QtW.QWidget()
        layout = QtW.QGridLayout(central)
        layout.setContentsMargins(2, 2, 2, 2)
        self._layout = layout
        self._toggle_switches: list[QChannelToggleSwitch] = []
        self.setWidget(central)
        self._label_font = QtGui.QFont(DefaultFontFamily, 8)

    def set_channels(self, channels: list[ChannelInfo]):
        labels = [ch.name for ch in channels]
        for ith in range(len(self._toggle_switches), len(labels)):
            sw = QChannelToggleSwitch(channels[ith])
            sw.setChecked(True)
            sw.setFont(self._label_font)
            sw.toggled.connect(self._emit_state_changed)
            row, col = divmod(ith, 3)
            self._layout.addWidget(sw, row, col)
            self._toggle_switches.append(sw)
        while len(self._toggle_switches) > len(labels):
            sw = self._toggle_switches.pop()
            sw.setParent(None)
        for i in range(len(channels)):
            sw = self._toggle_switches[i]
            sw.set_channel(channels[i])
        self.setFixedWidth(70 * min(3, len(self._toggle_switches)) + 10)

    def _emit_state_changed(self):
        self.stateChanged.emit()

    def _check_states(self) -> list[bool]:
        return [sw.isChecked() for sw in self._toggle_switches]



from himena._app_model._context import AppContext
from himena._app_model._application import get_model_app, HimenaApplication
from himena._app_model._command_registry import CommandsRegistry

__all__ = ["AppContext", "get_model_app", "CommandsRegistry", "HimenaApplication"]



from __future__ import annotations

from typing import Iterator, overload
from pathlib import Path

from himena.widgets import MainWindow
from contextlib import contextmanager


@overload
@contextmanager
def file_dialog_response(
    ui: MainWindow,
    path: str | Path,
    *,
    confirm: bool = False,
) -> Iterator[Path]: ...


@overload
@contextmanager
def file_dialog_response(
    ui: MainWindow,
    path: list[str | Path],
    *,
    confirm: bool = False,
) -> Iterator[list[Path]]: ...


@contextmanager
def file_dialog_response(
    ui: MainWindow,
    path: str | Path | list[str | Path],
    *,
    confirm: bool = False,
):
    """Set the response of the file dialog in this context."""
    old_inst = ui._instructions
    if isinstance(path, list):
        path = [Path(p) for p in path]
    else:
        path = Path(path)
    try:
        ui._instructions = ui._instructions.updated(
            file_dialog_response=lambda: path,
            confirm=confirm,
        )
        yield path
    finally:
        ui._instructions = old_inst


@contextmanager
def choose_one_dialog_response(ui: MainWindow, choice: str):
    """Set the response of the choose one dialog in this context."""
    old_inst = ui._instructions
    try:
        ui._instructions = ui._instructions.updated(
            choose_one_dialog_response=lambda: choice
        )
        yield
    finally:
        ui._instructions = old_inst



from __future__ import annotations

from uuid import UUID
from datetime import timedelta
from functools import singledispatch

from cmap import Color
from qtpy import QtWidgets as QtW, QtGui
from magicgui.widgets import request_values
from himena import workflow as _wf
from himena.consts import StandardType
from himena.plugins import validate_protocol
from himena.qt._utils import get_main_window
from himena.qt._qflowchart import QFlowChartWidget, BaseNodeItem
from himena.widgets import current_instance, MainWindow
from himena.style import Theme
from himena.types import WidgetDataModel
from himena.workflow._graph import _make_mock_main_window


class WorkflowNodeItem(BaseNodeItem):
    def __init__(self, step: _wf.WorkflowStep, main: MainWindow):
        self._item = _step_to_item(step, main)
        _add_common_child(self._item, step)
        self._step = step

    def text(self) -> str:
        return self._item.text(0)

    def color(self) -> Color:
        if self._step.process_output:
            return Color("#A3F020")
        if (
            isinstance(self._step, _wf.CommandExecution)
            and self._step.execution_time <= 0
        ):
            # parametric window
            return Color("#FFA2A2")
        return Color("#A2A3F0")

    def tooltip(self) -> str:
        return self._item.toolTip(0)

    def content(self) -> str:
        texts = []
        for i in range(self._item.childCount()):
            child = self._item.child(i)
            if child.text(0):
                texts.append(child.text(0))
            for j in range(child.childCount()):
                subchild = child.child(j)
                if subchild.text(0):
                    texts.append(f"  {subchild.text(0)}")
        return "\n".join(texts)

    def id(self):
        return self._step.id


class QWorkflowViewBase(QFlowChartWidget):
    @validate_protocol
    def theme_changed_callback(self, theme: Theme) -> None:
        self.setBackgroundBrush(QtGui.QColor(Color(theme.background).hex))

    def add_workflow(self, workflow: _wf.Workflow) -> None:
        try:
            main = current_instance()
        except StopIteration:
            main = _make_mock_main_window()
        for step in workflow:
            if step.id not in self.view._node_map:
                self._add_step(step, workflow, main)

    def clear_workflow(self) -> None:
        """Clear the workflow view."""
        self.scene.clear()
        self.view._node_map.clear()

    def _add_step(self, step: _wf.WorkflowStep, workflow: _wf.Workflow, main) -> None:
        parents: list[UUID] = []
        for _id in step.iter_parents():
            if _id not in self.view._node_map:
                parent_step = workflow.step_for_id(_id)
                self._add_step(parent_step, workflow, main)
            parents.append(_id)
        item = WorkflowNodeItem(step, main)
        self.view.add_child(item, parents=parents)


class QWorkflowView(QWorkflowViewBase):
    def __init__(self):
        super().__init__()
        self.view.item_right_clicked.connect(self._on_right_clicked)
        self._modified = False
        self._editable = True
        self._workflow: _wf.Workflow = _wf.Workflow()

    def set_workflow(self, workflow: _wf.Workflow) -> None:
        """Set the workflow."""
        self.clear_workflow()
        self._workflow = workflow
        self.add_workflow(workflow)

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        wf = model.value
        if not isinstance(wf, _wf.Workflow):
            raise ValueError(f"Expected Workflow, got {type(wf)}")
        self.set_workflow(wf)

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        return WidgetDataModel(
            value=self._workflow,
            type=self.model_type(),
            extension_default=".workflow.json",
        )

    @validate_protocol
    def model_type(self) -> str:
        if any(isinstance(step, _wf.UserInput) for step in self._workflow.steps):
            return StandardType.WORKFLOW_PARAMETRIC
        return StandardType.WORKFLOW

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 400, 420

    @validate_protocol
    def is_modified(self) -> bool:
        return self._modified

    @validate_protocol
    def set_modified(self, modified: bool) -> None:
        self._modified = modified

    @validate_protocol
    def is_editable(self) -> bool:
        return self._editable

    @validate_protocol
    def set_editable(self, editable: bool) -> None:
        self._editable = editable

    def _replace_with_file_reader(self, item: WorkflowNodeItem, how: str) -> None:
        wf_new = self._workflow.replace_with_input(item.id(), how=how)
        self.set_workflow(wf_new)
        self._modified = True

    def _toggle_to_be_added(self, item: WorkflowNodeItem) -> None:
        step = item._step
        step.process_output = not step.process_output
        self.set_workflow(self._workflow)
        self._modified = True

    def _find_window(self, item: WorkflowNodeItem) -> None:
        step = item._step
        main = get_main_window(self)
        for i_tab, tab in main.tabs.enumerate():
            for i_win, win in tab.enumerate():
                if win._identifier == step.id:
                    main.tabs.current_index = i_tab
                    main.tabs[i_tab].current_index = i_win
                    return None
        raise ValueError("No window in the main window matches the workflow step.")

    def _execute_upto(self, item: WorkflowNodeItem) -> None:
        wf = self._workflow
        wf.filter(item.id()).compute(process_output=True)

    def _edit_user_input(self, item: WorkflowNodeItem) -> None:
        wf = self._workflow
        step = item._step
        if not isinstance(step, _wf.UserInput):
            raise ValueError(f"Expected UserInput, got {type(step)}")
        resp = request_values(
            label={"value": step.label, "label": "Label"},
            doc={"value": step.doc, "label": "Docstring"},
            title="Edit User Input",
            parent=self,
        )
        if resp:
            step.label = resp["label"]
            step.doc = resp["doc"]
            self.set_workflow(wf)  # update
            self._modified = True

    def _on_right_clicked(self, item: WorkflowNodeItem) -> None:
        pos = QtGui.QCursor.pos()
        self._make_context_menu(item).exec(pos)

    def _make_context_menu(self, item: WorkflowNodeItem) -> QtW.QMenu:
        menu = QtW.QMenu(self)
        menu.setToolTipsVisible(True)
        step = item._step
        a0 = menu.addAction(
            "Replace with file reader",
            lambda: self._replace_with_file_reader(item, "file"),
        )
        a0.setToolTip("Replace the selected item with a file reader")
        a1 = menu.addAction(
            "Replace with model input",
            lambda: self._replace_with_file_reader(item, "model"),
        )
        a1.setToolTip("Replace the selected item with a model input")
        a2 = menu.addAction("To be added", lambda: self._toggle_to_be_added(item))
        a2.setCheckable(True)
        a2.setChecked(step.process_output)
        a2.setToolTip(
            "Mark the selected item to add the output to the main window after the \n"
            "workflow execution (even if it's an intermediate step)"
        )
        menu.addSeparator()
        a3 = menu.addAction("Execute upto here", lambda: self._execute_upto(item))
        a3.setToolTip("Execute the workflow up to the selected item")
        a4 = menu.addAction("Find window", lambda: self._find_window(item))
        a4.setToolTip("Find the window that corresponds to the selected item")
        a5 = menu.addAction("Edit ...", lambda: self._edit_user_input(item))
        a5.setToolTip("Edit the selected user input item")

        # update enabled state
        a0.setEnabled(self.is_editable())
        a1.setEnabled(self.is_editable())
        a2.setEnabled(self.is_editable())
        a5.setEnabled(self.is_editable())
        if isinstance(step, _wf.UserInput):
            a3.setEnabled(False)
        else:
            a5.setEnabled(False)
        if item.id() == self._workflow.last_id():
            a2.setEnabled(False)
        return menu


@singledispatch
def _step_to_item(step: _wf.WorkflowStep, main: MainWindow) -> QtW.QTreeWidgetItem:
    raise ValueError(f"Unknown workflow node type {type(step)}")


@_step_to_item.register
def _(step: _wf.LocalReaderMethod, main: MainWindow) -> QtW.QTreeWidgetItem:
    item = QtW.QTreeWidgetItem([f"[Local File]\ntype={step.output_model_type!r}"])
    if isinstance(step.path, list):
        for i, path in enumerate(step.path):
            item.addChild(QtW.QTreeWidgetItem([f"({i}) {path.as_posix()}"]))
    else:
        item.addChild(QtW.QTreeWidgetItem([f"{step.path.as_posix()}"]))
    item.addChild(QtW.QTreeWidgetItem([f"plugin = {step.plugin!r}"]))
    item.setToolTip(0, str(step.path))
    return item


@_step_to_item.register
def _(step: _wf.RemoteReaderMethod, main: MainWindow) -> QtW.QTreeWidgetItem:
    item = QtW.QTreeWidgetItem([f"[Remote File]\ntype={step.output_model_type!r}"])
    item.addChild(QtW.QTreeWidgetItem([f"{step.to_str()}"]))
    item.addChild(QtW.QTreeWidgetItem([f"plugin = {step.plugin!r}"]))
    item.setToolTip(0, str(step.path))
    return item


@_step_to_item.register
def _(step: _wf.UserModification, main: MainWindow) -> QtW.QTreeWidgetItem:
    item = QtW.QTreeWidgetItem(["[User Modification]"])
    return item


@_step_to_item.register
def _(step: _wf.UserInput, main: MainWindow) -> QtW.QTreeWidgetItem:
    if step.how == "file":
        text = "File Input"
    elif step.how == "model":
        text = "Model Input"
    else:
        text = "User Input"
    item = QtW.QTreeWidgetItem([f"[{text}]"])
    item.addChild(QtW.QTreeWidgetItem([f"label = {step.label!r}"]))
    item.addChild(QtW.QTreeWidgetItem([f"doc = {step.doc!r}"]))
    item.setToolTip(0, f"{text}\nlabel = {step.label}\ndoc = {step.doc}")
    return item


@_step_to_item.register
def _(step: _wf.CommandExecution, main: MainWindow) -> QtW.QTreeWidgetItem:
    if action := main.model_app.registered_actions.get(step.command_id, None):
        title = action.title
    else:
        title = step.command_id
    item = QtW.QTreeWidgetItem([f"[Command]\n{title}"])
    item.setToolTip(0, step.command_id)
    item.addChild(QtW.QTreeWidgetItem([f"command_id = {step.command_id!r}"]))
    for param in step.parameters or []:
        if isinstance(param, _wf.UserParameter):
            child = QtW.QTreeWidgetItem([f"(parameter) {param.name} = {param.value!r}"])
        elif isinstance(param, _wf.ModelParameter):
            child = QtW.QTreeWidgetItem(
                [f"(parameter) {param.name} = <data model, type={param.model_type!r}>"]
            )
        elif isinstance(param, _wf.WindowParameter):
            child = QtW.QTreeWidgetItem(
                [f"(parameter) {param.name} = <window, type={param.model_type!r}>"]
            )
        elif isinstance(param, _wf.ListOfModelParameter):
            short_desc = "<subwindows>" if param.is_window else "<models>"
            child = QtW.QTreeWidgetItem([f"(parameter) {param.name} = {short_desc}"])
        else:
            raise ValueError(f"Unknown parameter type {type(param)}")
        item.addChild(child)
    for ctx in step.contexts:
        if isinstance(ctx, _wf.ModelParameter):
            child = QtW.QTreeWidgetItem(
                [f"(context) <data model, type={ctx.model_type!r}>"]
            )
        elif isinstance(ctx, _wf.WindowParameter):
            child = QtW.QTreeWidgetItem(
                [f"(context) <window, type={ctx.model_type!r}>"]
            )
        else:
            raise ValueError(f"Unknown context type {type(ctx)}")
        item.addChild(child)
    if (dt := step.execution_time) > 0.0:
        item.addChild(
            QtW.QTreeWidgetItem([f"execution_time = {timedelta(seconds=dt)}"])
        )
    return item


@_step_to_item.register
def _(step: _wf.ProgrammaticMethod, main: MainWindow) -> QtW.QTreeWidgetItem:
    item = QtW.QTreeWidgetItem(
        [f"[Programmatic Method]\ntype={step.output_model_type!r}"]
    )
    return item


def _add_common_child(item: QtW.QTreeWidgetItem, step: _wf.WorkflowStep):
    item.addChild(
        QtW.QTreeWidgetItem([f"datetime = {step.datetime:%Y-%m-%d %H:%M:%S}"])
    )
    return item



"""Test consistency and completeness of frontend widget implementations.

This submodule provides a set of helper functions to test the widget implementations.
Note that they are not meant to test the GUI, but the logic of `update_model` and
`to_model` methods of the widgets.
"""

from himena.testing import image, table
from himena.testing.subwindow import WidgetTester
from himena.testing._pytest import install_plugin
from himena.testing.dialog import file_dialog_response, choose_one_dialog_response

__all__ = [
    "image",
    "table",
    "WidgetTester",
    "install_plugin",
    "file_dialog_response",
    "choose_one_dialog_response",
]



from __future__ import annotations

from functools import partial
from logging import getLogger
from typing import TYPE_CHECKING, Iterable, Literal
import uuid
import weakref
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt
from superqt import QElidingLabel
from himena.types import WidgetDataModel
from himena.utils.misc import is_subtype
from himena.qt._qsub_window import QSubWindow, QSubWindowArea, get_subwindow
from himena.qt._utils import get_main_window
from himena import _drag

if TYPE_CHECKING:
    from himena.widgets import SubWindow

_LOGGER = getLogger(__name__)
_NONE_TOOLTIP = "Drop a subwindow here by Ctrl+dragging the title bar."


class QModelDropBase(QtW.QGroupBox):
    close_requested = QtCore.Signal(object)  # emit self

    def __init__(
        self, layout: Literal["horizontal", "vertical"] = "horizontal", parent=None
    ):
        super().__init__(parent)
        self._thumbnail = _QImageLabel()
        self._target_id: uuid.UUID | None = None
        self._data_model: WidgetDataModel | None = None
        self._main_window_ref = lambda: None
        self._label = QElidingLabel()
        self._label.setAlignment(
            Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
        )
        if layout == "horizontal":
            self._label.setFixedHeight(THUMBNAIL_SIZE.height() + 2)
            _layout = QtW.QHBoxLayout(self)
            _layout.setAlignment(
                Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop
            )
            self._label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        else:
            self._label.setMinimumWidth(150)
            _layout = QtW.QVBoxLayout(self)
            _layout.setAlignment(
                Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignLeft
            )
            self._label.setAlignment(
                Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
            )
        self._label.setToolTip(_NONE_TOOLTIP)
        _layout.setContentsMargins(1, 1, 1, 1)
        _layout.addWidget(self._thumbnail)
        _layout.addWidget(self._label)

        self._close_btn = QtW.QToolButton()
        self._close_btn.setCursor(Qt.CursorShape.PointingHandCursor)
        self._close_btn.setText("✕")
        self._close_btn.setFixedSize(15, 15)
        self._close_btn.clicked.connect(lambda: self.close_requested.emit(self))
        self._close_btn.setParent(self)
        self._close_btn.hide()

    def _update_btn_pos(self):
        pos_loc = self.rect().topRight() - QtCore.QPoint(
            self._close_btn.width() + 5, -5
        )
        self._close_btn.move(pos_loc)

    def enterEvent(self, a0):
        self._close_btn.show()
        self._update_btn_pos()
        return super().enterEvent(a0)

    def leaveEvent(self, a0):
        rect = self.rect().adjusted(1, 1, -1, -1)
        if not rect.contains(self.mapFromGlobal(QtGui.QCursor.pos())):
            self._close_btn.hide()
        return super().leaveEvent(a0)

    def moveEvent(self, a0):
        self._update_btn_pos()
        return super().moveEvent(a0)

    def resizeEvent(self, a0):
        self._update_btn_pos()
        return super().resizeEvent(a0)

    def subwindow(self) -> SubWindow | None:
        """The dropped subwindow."""
        if self._target_id is None:
            return None
        ui = self._main_window_ref()
        if ui is None:
            return None
        return ui.window_for_id(self._target_id)

    def to_model(self):
        if self._data_model is not None:
            return self._data_model
        if widget := self.subwindow():
            return widget.to_model()
        return None

    def set_qsubwindow(self, src: QSubWindow):
        src_wrapper = src._my_wrapper()
        self._thumbnail.set_pixmap(
            src._pixmap_resized(THUMBNAIL_SIZE, QtGui.QColor("#f0f0f0"))
        )
        self._target_id = src_wrapper._identifier
        self._main_window_ref = weakref.ref(get_main_window(src))
        self._label.setText(src.windowTitle())

    def set_subwindow(self, src: SubWindow | uuid.UUID):
        if isinstance(src, uuid.UUID):
            src = get_main_window(self).window_for_id(src)
        self.set_qsubwindow(get_subwindow(src.widget))


class QModelDrop(QModelDropBase):
    """Widget for dropping model data from a subwindow."""

    valueChanged = QtCore.Signal(WidgetDataModel)
    windowChanged = QtCore.Signal(object)

    def __init__(
        self,
        types: list[str] | None = None,
        layout: Literal["horizontal", "vertical"] = "horizontal",
        parent: QtW.QWidget | None = None,
    ):
        super().__init__(layout, parent)
        self.setAcceptDrops(True)
        self._label.setText("Drop here")
        self._label.setAlignment(
            Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter
        )
        self._allowed_types = types  # the model type
        self.close_requested.connect(lambda: self.set_model(None))

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(150, 50)

    def enterEvent(self, a0):
        if self._data_model is None and self._target_id is None:
            return None
        return super().enterEvent(a0)

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent):
        if isinstance(src := event.source(), QSubWindow):
            widget = src._widget
            if not hasattr(widget, "to_model"):
                _LOGGER.debug("Ignoring drop event")
                event.ignore()
                event.setDropAction(Qt.DropAction.IgnoreAction)
                return
            model_type = getattr(widget, "model_type", lambda: None)()
            _LOGGER.info("Entered model type %s", model_type)
            if self._is_type_maches(model_type):
                _LOGGER.debug("Accepting drop event")
                event.accept()
                return
        elif isinstance(area := event.source(), QSubWindowArea):
            subwindows = area.subWindowList()
            if len(subwindows) == 1:
                event.accept()
                return
        elif model := _drag.get_dragging_model():
            if self._is_type_maches(model.type):
                _LOGGER.debug("Accepting drop event")
                event.accept()
                return
        event.ignore()
        event.setDropAction(Qt.DropAction.IgnoreAction)

    def dropEvent(self, event: QtGui.QDropEvent):
        if isinstance(win := event.source(), QSubWindow):
            self._drop_qsubwindow(win)
        elif isinstance(area := event.source(), QSubWindowArea):
            subwindows = area.subWindowList()
            if len(subwindows) == 1:
                self._drop_qsubwindow(subwindows[0])
        elif model := _drag.get_dragging_model():
            self.set_model(model.data_model())

    def set_model(self, value: WidgetDataModel | uuid.UUID | None):
        if value is None:
            self._label.setText("Drop here")
            self._label.setToolTip(_NONE_TOOLTIP)
            self._thumbnail.unset_pixmap()
        else:
            if isinstance(value, uuid.UUID):
                return self.set_subwindow(value)
            self._data_model = value
            self._label.setText(f"✓ {value.value!r}")
            self._label.setToolTip(repr(value))

    def _drop_qsubwindow(self, win: QSubWindow):
        widget = win._widget
        model_type = getattr(widget, "model_type", lambda: None)()
        _LOGGER.info("Dropped model type %s", model_type)
        if self._is_type_maches(model_type):
            _LOGGER.info("Dropped model %s", win.windowTitle())
            self.set_qsubwindow(win)
            self._emit_window(win._my_wrapper())
            # move this window to the top
            this = self
            while this := this.parent():
                if isinstance(this, QSubWindow):
                    main = get_main_window(this)
                    main.current_window = this._my_wrapper()
                    break

    def _emit_window(self, win: SubWindow):
        self.windowChanged.emit(win)
        if win.supports_to_model:
            self.valueChanged.emit(win.to_model())

    def _is_type_maches(self, model_type: str) -> bool:
        if self._allowed_types is None:
            return True
        return any(is_subtype(model_type, t) for t in self._allowed_types)


class QModelDropList(QtW.QListWidget):
    modelsChanged = QtCore.Signal(list)
    windowsChanged = QtCore.Signal(list)

    def __init__(
        self,
        types: list[str] | None = None,
        layout: Literal["horizontal", "vertical"] = "vertical",
        parent: QtW.QWidget | None = None,
    ):
        super().__init__(parent)
        if layout == "horizontal":
            self.setFlow(QtW.QListView.Flow.LeftToRight)
        else:
            self.setFlow(QtW.QListView.Flow.TopToBottom)
        self.setResizeMode(QtW.QListView.ResizeMode.Adjust)
        self.setAcceptDrops(True)
        self._allowed_types = types  # the model type
        self._layout = layout

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(250, 200)

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent):
        if isinstance(src := event.source(), QSubWindow):
            widget = src._widget
            if not hasattr(widget, "to_model"):
                _LOGGER.debug("Ignoring drop event")
                event.ignore()
                event.setDropAction(Qt.DropAction.IgnoreAction)
                return
            model_type = getattr(widget, "model_type", lambda: None)()
            _LOGGER.info("Entered model type: %s", model_type)
            if self._is_type_maches(model_type):
                _LOGGER.debug("Accepting drop event")
                event.accept()
                event.setDropAction(Qt.DropAction.MoveAction)
                return
        elif isinstance(area := event.source(), QSubWindowArea):
            subwindows = area.subWindowList()
            if len(subwindows) == 1:
                event.accept()
                return
        elif model := _drag.get_dragging_model():
            if self._is_type_maches(model.type):
                _LOGGER.debug("Accepting drop event")
                event.accept()
                event.setDropAction(Qt.DropAction.MoveAction)
                return
        event.ignore()
        event.setDropAction(Qt.DropAction.IgnoreAction)

    def dragMoveEvent(self, e: QtGui.QDragMoveEvent):
        e.acceptProposedAction()
        return

    def dropEvent(self, event: QtGui.QDropEvent):
        if isinstance(win := event.source(), QSubWindow):
            self._drop_qsubwindow(win)
            event.accept()
            return
        elif isinstance(area := event.source(), QSubWindowArea):
            subwindows = area.subWindowList()
            if len(subwindows) == 1:
                self._drop_qsubwindow(subwindows[0])
                event.accept()
                return
        event.ignore()
        event.setDropAction(Qt.DropAction.IgnoreAction)
        return None

    def leaveEvent(self, a0):
        for i in range(self.count()):
            item = self.item(i)
            widget = self.itemWidget(item)
            widget.leaveEvent(a0)
        return super().leaveEvent(a0)

    def _is_type_maches(self, model_type: str) -> bool:
        if self._allowed_types is None:
            return True
        return any(is_subtype(model_type, t) for t in self._allowed_types)

    def _drop_qsubwindow(self, win: QSubWindow):
        widget = win._widget
        model_type = getattr(widget, "model_type", lambda: None)()
        _LOGGER.info("Dropped model type %s", model_type)
        if self._is_type_maches(model_type):
            _LOGGER.info("Dropped model %s", win.windowTitle())
            self._append_sub_window(win)
            self.windowsChanged.emit(self.windows())
            self.modelsChanged.emit(self.models())

    def _append_item(self) -> QModelListItem:
        item = QtW.QListWidgetItem()
        self.addItem(item)
        item.setFlags(Qt.ItemFlag.ItemIsEnabled)
        if self.flow() == QtW.QListView.Flow.LeftToRight:
            item_widget = QModelListItem(layout="vertical")
            item.setSizeHint(QtCore.QSize(100, 200))
        else:
            item_widget = QModelListItem(layout="horizontal")
            item.setSizeHint(QtCore.QSize(100, THUMBNAIL_SIZE.height() + 2))
        self.setItemWidget(item, item_widget)
        item_widget.close_requested.connect(self._remove_item)
        return item_widget

    def _append_sub_window(self, src: QSubWindow):
        item_widget = self._append_item()
        item_widget.set_qsubwindow(src)
        win = src._my_wrapper()
        win.closed.connect(partial(self._remove_item, item_widget))

    def _remove_item(self, item: QModelListItem):
        for i in range(self.count()):
            if self.itemWidget(self.item(i)) is item:
                if (win := item.subwindow()) and (
                    cb := partial(self._remove_item, item)
                ) in win.closed:
                    # disconnect the signal to avoid memory leak
                    win.closed.disconnect(cb)
                self.takeItem(i)
                return

    def models(self) -> list[WidgetDataModel]:
        """List of models."""
        return [self.itemWidget(self.item(i)).to_model() for i in range(self.count())]

    def set_models(self, value):
        if value is None:
            self.clear()
        else:
            raise ValueError("Cannot set list of WidgetDataModel directly.")

    def windows(self) -> list[SubWindow]:
        """List of subwindows."""
        return [self.itemWidget(self.item(i)).subwindow() for i in range(self.count())]

    def set_windows(self, value: Iterable[SubWindow] | None):
        self.clear()
        if value is not None:
            for win in value:
                qwin = get_subwindow(win.widget)
                self._append_sub_window(qwin)

    if TYPE_CHECKING:

        def itemWidget(self, item: QtW.QListWidgetItem) -> QModelListItem: ...


class QModelListItem(QModelDropBase):
    pass


THUMBNAIL_SIZE = QtCore.QSize(36, 36)


class _QImageLabel(QtW.QLabel):
    def __init__(self):
        super().__init__()
        self.setAlignment(Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignTop)
        self.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )
        self.setFixedSize(0, 0)

    def set_pixmap(self, pixmap: QtGui.QPixmap):
        self.setFixedSize(THUMBNAIL_SIZE)
        sz = self.size()
        self.setPixmap(
            pixmap.scaled(
                sz,
                Qt.AspectRatioMode.KeepAspectRatio,
                Qt.TransformationMode.SmoothTransformation,
            )
        )

    def unset_pixmap(self):
        self.setFixedSize(0, 0)
        self.clear()



from __future__ import annotations

import logging
import numpy as np

from himena import MainWindow
from himena.consts import MonospaceFontFamily, StandardType
from himena.standards.model_meta import ImageMeta

from qtpy import QtWidgets as QtW, QtCore, QtGui

_LOGGER = logging.getLogger(__name__)


class QImageDrawToolWidget(QtW.QWidget):
    def __init__(self, ui: MainWindow):
        self._ui = ui
        super().__init__()
        layout = QtW.QVBoxLayout(self)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        self._fill_btn = QtW.QPushButton("Fill")
        self._fill_value = QtW.QLineEdit()
        self._fill_value.setFont(QtGui.QFont(MonospaceFontFamily))
        layout.addWidget(self._fill_btn)
        layout.addWidget(self._fill_value)

        self._fill_btn.clicked.connect(self._fill)

    def _fill(self):
        win = self._ui.current_window
        if win is None:
            _LOGGER.info("No window is selected.")
            return
        if not win.is_editable:
            _LOGGER.info("Window is not editable.")
            return
        model = win.to_model()
        if model.type != StandardType.IMAGE:
            _LOGGER.info("Model is not an image.")
            return
        if not isinstance(meta := model.metadata, ImageMeta):
            _LOGGER.info("metadata is not ImageMeta")
            return
        if meta.is_rgb:
            raise NotImplementedError("RGB image is not supported yet.")
        roi = meta.current_roi
        if roi is None:
            _LOGGER.info("No ROI is selected.")
            return
        if not isinstance(arr := model.value, np.ndarray):
            raise ValueError("Model value is not a numpy array.")
        mask = roi.to_mask(arr.shape)
        dtype = np.dtype(arr.dtype)
        arr[mask] = dtype.type(self._fill_value.text())
        win.update_value(arr)



from __future__ import annotations

from pathlib import Path
from himena.plugins import register_reader_plugin, register_writer_plugin
from himena.data_wrappers import list_installed_dataframe_packages
from himena.types import WidgetDataModel
from himena_builtins import _io
from himena.consts import (
    StandardType,
    BasicTextFileTypes,
    BasicImageFileTypes,
    ConventionalTextFileNames,
    ExcelFileTypes,
)


@register_reader_plugin(priority=50)
def read_text(file_path: Path) -> WidgetDataModel:
    suffix = file_path.suffix.rstrip("~")
    if suffix == ".csv":
        return _io.default_csv_reader(file_path)
    elif suffix == ".tsv":
        return _io.default_tsv_reader(file_path)
    elif suffix in BasicImageFileTypes:
        return _io.default_image_reader(file_path)
    elif suffix == ".json":
        if len(suffixes := file_path.suffixes) > 1:
            if suffixes[-2:] == [".plot", ".json"]:
                return _io.default_plot_reader(file_path)
            elif suffixes[-2:] == [".roi", ".json"]:
                return _io.default_roi_reader(file_path)
            elif suffixes[-2:] == [".workflow", ".json"]:
                return _io.default_workflow_reader(file_path)
        return _io.default_text_reader(file_path)
    elif suffix in BasicTextFileTypes:
        return _io.default_text_reader(file_path)
    elif file_path.name.rstrip("~") in ConventionalTextFileNames:
        return _io.default_text_reader(file_path)
    raise ValueError(f"Unsupported file type: {file_path.suffix}")


@read_text.define_matcher
def _(file_path: Path) -> str | None:
    suffix = file_path.suffix.rstrip("~")
    if suffix == ".csv":
        return StandardType.TABLE
    elif suffix == ".tsv":
        return StandardType.TABLE
    elif suffix in BasicImageFileTypes:
        return StandardType.IMAGE
    elif suffix == ".json":
        if len(suffixes := file_path.suffixes) > 1:
            if suffixes[-2:] == [".plot", ".json"]:
                return StandardType.PLOT
            elif suffixes[-2:] == [".roi", ".json"]:
                return StandardType.ROIS
            elif suffixes[-2:] == [".workflow", ".json"]:
                return StandardType.WORKFLOW
        return StandardType.TEXT
    elif suffix in BasicTextFileTypes:
        return StandardType.TEXT
    elif file_path.name.rstrip("~") in ConventionalTextFileNames:
        return StandardType.TEXT
    return None


@register_reader_plugin(priority=50)
def read_file_list(file_path: Path | list[Path]) -> WidgetDataModel:
    return _io.default_file_list_reader(file_path)


@read_file_list.define_matcher
def _(file_path: Path | list[Path]) -> str | None:
    if isinstance(file_path, list) or file_path.is_dir():
        return StandardType.MODELS
    return None


@register_reader_plugin(priority=50)
def read_image(file_path: Path) -> WidgetDataModel:
    if file_path.suffix.rstrip("~") in {".png", ".jpg", ".jpeg"}:
        return _io.default_image_reader(file_path)
    raise ValueError(f"Unsupported file type: {file_path.suffix}")


@read_image.define_matcher
def _(file_path: Path) -> str | None:
    if file_path.suffix.rstrip("~") in {".png", ".jpg", ".jpeg"}:
        return StandardType.IMAGE
    return None


@register_reader_plugin(priority=50)
def read_excel(file_path: Path) -> WidgetDataModel:
    if file_path.suffix.rstrip("~") in ExcelFileTypes:
        return _io.default_excel_reader(file_path)
    raise ValueError(f"Unsupported file type: {file_path.suffix}")


@read_excel.define_matcher
def _(file_path: Path) -> str | None:
    if file_path.suffix.rstrip("~") in ExcelFileTypes:
        return StandardType.EXCEL
    return None


@register_reader_plugin(priority=50)
def read_numpy_array(file_path: Path) -> WidgetDataModel:
    if file_path.suffix.rstrip("~") == ".npy":
        return _io.default_array_reader(file_path)
    raise ValueError(f"Unsupported file type: {file_path.suffix}")


@read_numpy_array.define_matcher
def _(file_path: Path) -> str | None:
    if file_path.suffix.rstrip("~") == ".npy":
        return StandardType.ARRAY
    return None


@register_reader_plugin(priority=50)
def read_pickle(file_path: Path) -> WidgetDataModel:
    if file_path.suffix.rstrip("~") == ".pickle":
        return _io.default_pickle_reader(file_path)
    raise ValueError(f"Unsupported file type: {file_path.suffix}")


@read_pickle.define_matcher
def _(file_path: Path) -> str | None:
    if file_path.suffix.rstrip("~") == ".pickle":
        return StandardType.ANY
    return None


@register_reader_plugin(priority=50)
def read_zip(file_path: Path) -> WidgetDataModel:
    if file_path.suffix.rstrip("~") == ".zip":
        return _io.default_zip_reader(file_path)
    raise ValueError(f"Unsupported file type: {file_path.suffix}")


@read_zip.define_matcher
def _(file_path: Path) -> str | None:
    if file_path.suffix.rstrip("~") == ".zip":
        return StandardType.MODELS
    return None


@register_reader_plugin(priority=-100)
def read_as_text_anyway(file_path: Path) -> WidgetDataModel:
    return _io.default_plain_text_reader(file_path)


@read_as_text_anyway.define_matcher
def _(file_path: Path) -> str | None:
    return StandardType.TEXT


@register_reader_plugin(priority=0)
def read_as_unknown(file_path: Path) -> WidgetDataModel:
    return _io.fallback_reader(file_path)


@read_as_unknown.define_matcher
def _(file_path: Path) -> str | None:
    return StandardType.READER_NOT_FOUND


@register_reader_plugin(priority=50)
def read_as_pandas_dataframe(file_path: Path) -> WidgetDataModel:
    suffix = file_path.suffix.rstrip("~")
    if suffix in {".csv", ".txt"}:
        return _io.DataFrameReader("pandas", "read_csv", {})(file_path)
    elif suffix == ".tsv":
        return _io.DataFrameReader("pandas", "read_csv", {"sep": "\t"})(file_path)
    elif suffix in {".html", ".htm"}:
        return _io.DataFrameReader("pandas", "read_html", {})(file_path)
    elif suffix == ".json":
        return _io.DataFrameReader("pandas", "read_json", {})(file_path)
    elif suffix in {".parquet", ".pq"}:
        return _io.DataFrameReader("pandas", "read_parquet", {})(file_path)
    elif suffix == ".feather":
        return _io.DataFrameReader("pandas", "read_feather", {})(file_path)
    raise ValueError(f"Unsupported file type: {file_path.suffix}")


@register_reader_plugin(priority=50)
def read_as_polars_dataframe(file_path: Path) -> WidgetDataModel:
    suffix = file_path.suffix.rstrip("~")
    if suffix in {".csv", ".txt"}:
        return _io.DataFrameReader("polars", "read_csv", {})(file_path)
    elif suffix == ".tsv":
        return _io.DataFrameReader("polars", "read_csv", {"separator": "\t"})(file_path)
    elif suffix == ".feather":
        return _io.DataFrameReader("polars", "read_ipc", {})(file_path)
    elif suffix == ".json":
        return _io.DataFrameReader("polars", "read_json", {})(file_path)
    elif suffix in {".parquet", ".pq"}:
        return _io.DataFrameReader("polars", "read_parquet", {})(file_path)
    raise ValueError(f"Unsupported file type: {suffix}")


@register_reader_plugin(priority=20)
def read_as_pandas_plot(file_path: Path) -> WidgetDataModel:
    model = read_as_pandas_dataframe(file_path)
    model.type = StandardType.DATAFRAME_PLOT
    return model


@register_reader_plugin(priority=20)
def read_as_polars_plot(file_path: Path) -> WidgetDataModel:
    model = read_as_polars_dataframe(file_path)
    model.type = StandardType.DATAFRAME_PLOT
    return model


@read_as_pandas_dataframe.define_matcher
@read_as_pandas_plot.define_matcher
def _(file_path: Path) -> str | None:
    if "pandas" not in list_installed_dataframe_packages():
        return None
    if file_path.suffix.rstrip("~") in {
        ".csv", ".txt", ".tsv", ".html", ".htm", ".json", ".parquet", ".pq", ".feather",
    }:  # fmt: skip
        return StandardType.DATAFRAME
    return None


@read_as_polars_dataframe.define_matcher
@read_as_polars_plot.define_matcher
def _(file_path: Path) -> str | None:
    if "polars" not in list_installed_dataframe_packages():
        return None
    if file_path.suffix.rstrip("~") in {
        ".csv", ".txt", ".tsv", ".feather", ".json", ".parquet", ".pq",
    }:  # fmt: skip
        return StandardType.DATAFRAME
    return None


@register_writer_plugin(priority=50)
def write_text(model: WidgetDataModel, path: Path):
    return _io.default_text_writer(model, path)


@write_text.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.TEXT) and isinstance(model.value, str)


@register_writer_plugin(priority=50)
def write_table(model: WidgetDataModel, path: Path):
    return _io.default_table_writer(model, path)


@write_table.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.TABLE)


@register_writer_plugin(priority=50)
def write_image(model: WidgetDataModel, path: Path):
    return _io.default_image_writer(model, path)


@write_image.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.IMAGE)


@register_writer_plugin(priority=10)
def write_dict(model: WidgetDataModel, path: Path):
    return _io.default_dict_writer(model, path)


@write_dict.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.DICT)


@register_writer_plugin(priority=50)
def write_excel(model: WidgetDataModel, path: Path):
    return _io.default_excel_writer(model, path)


@write_excel.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.EXCEL)


@register_writer_plugin(priority=50)
def write_array(model: WidgetDataModel, path: Path):
    return _io.default_array_writer(model, path)


@write_array.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.ARRAY)


@register_writer_plugin(priority=50)
def write_plot(model: WidgetDataModel, path: Path):
    return _io.default_plot_writer(model, path)


@write_plot.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.PLOT)


@register_writer_plugin(priority=50)
def write_roi(model: WidgetDataModel, path: Path):
    return _io.default_roi_writer(model, path)


@write_roi.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.ROIS)


@register_writer_plugin(priority=50)
def write_dataframe(model: WidgetDataModel, path: Path):
    return _io.default_dataframe_writer(model, path)


@write_dataframe.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.DATAFRAME)


@register_writer_plugin(priority=50)
def write_models(model: WidgetDataModel, path: Path):
    return _io.default_models_writer(model, path)


@write_models.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.MODELS)


@register_writer_plugin(priority=50)
def write_workflow(model: WidgetDataModel, path: Path):
    return _io.default_workflow_writer(model, path)


@write_workflow.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return model.is_subtype_of(StandardType.WORKFLOW)


@register_writer_plugin(priority=-50)
def write_pickle_anyway(model: WidgetDataModel, path: Path):
    return _io.default_pickle_writer(model, path)


@write_pickle_anyway.define_matcher
def _(model: WidgetDataModel, path: Path) -> bool:
    return True



from pathlib import Path

from qtpy import QtCore
import pytest
from magicgui.types import FileDialogMode
from himena.qt.magicgui import _file_edit
from himena.qt.magicgui._file_edit import FileEdit, QFileEdit

_directory = str(Path(__file__).parent)

def test_file_edit():
    fe = FileEdit()
    assert isinstance(fe.native, QFileEdit)
    fe.set_value(__file__)
    assert fe.value == Path(__file__)
    fe.set_value(None)
    assert fe.value is None
    with pytest.raises(ValueError):
        fe.set_value([__file__, Path(__file__).parent])
    fe.native._mode = FileDialogMode.EXISTING_FILES
    fe.set_value([__file__, Path(__file__).parent])
    assert fe.value == [Path(__file__), Path(__file__).parent]

def test_open_file_dialog(monkeypatch):
    fe = FileEdit()
    assert isinstance(fe.native, QFileEdit)
    fe.native._mode = FileDialogMode.EXISTING_FILE
    monkeypatch.setattr(_file_edit, "show_file_dialog", lambda *_, **__: __file__)
    fe.native._open_file_dialog()
    fe.native._mode = FileDialogMode.EXISTING_FILES
    monkeypatch.setattr(_file_edit, "show_file_dialog", lambda *_, **__: [__file__, _directory])
    fe.native._open_file_dialog()
    fe.native._mode = FileDialogMode.EXISTING_DIRECTORY
    monkeypatch.setattr(_file_edit, "show_file_dialog", lambda *_, **__: _directory)
    fe.native._open_file_dialog()
    fe.native._mode = FileDialogMode.OPTIONAL_FILE
    monkeypatch.setattr(_file_edit, "show_file_dialog", lambda *_, **__: __file__)
    fe.native._open_file_dialog()

def test_accept_urls():
    fe = FileEdit()
    assert isinstance(fe.native, QFileEdit)
    fe.native._mode = FileDialogMode.EXISTING_FILE
    fe.native._accept_urls([QtCore.QUrl(__file__)])
    fe.native._set_urls([QtCore.QUrl(__file__)])
    fe.native._mode = FileDialogMode.EXISTING_FILES
    fe.native._accept_urls([QtCore.QUrl(__file__), QtCore.QUrl(_directory)])
    fe.native._set_urls([QtCore.QUrl(__file__), QtCore.QUrl(_directory)])
    fe.native._mode = FileDialogMode.EXISTING_DIRECTORY
    fe.native._accept_urls([QtCore.QUrl(_directory)])
    fe.native._set_urls([QtCore.QUrl(_directory)])



from __future__ import annotations

from contextlib import contextmanager, suppress
from functools import wraps
import uuid
from typing import (
    Callable,
    Generic,
    Iterator,
    Sequence,
    TypeVar,
    overload,
    TYPE_CHECKING,
)
import weakref

from app_model.types import Action, ToggleRule

from himena.types import DockArea, DockAreaString
from himena.consts import NO_RECORDING_FIELD, MenuId
from himena.plugins.actions import (
    normalize_keybindings,
    AppActionRegistry,
    command_id_from_func,
    tooltip_from_func,
    norm_menus,
    PluginConfigTuple,
)

if TYPE_CHECKING:
    from typing import Self
    from himena.widgets import MainWindow, DockWidget
    from himena.widgets._wrapper import WidgetWrapper
    from himena.plugins.actions import KeyBindingsType, PluginConfigType

_F = TypeVar("_F", bound=Callable)
_W = TypeVar("_W", bound="WidgetWrapper")


@overload
def register_dock_widget_action(
    widget_factory: _F,
    *,
    menus: str | Sequence[str] | None = None,
    title: str | None = None,
    area: DockArea | DockAreaString = DockArea.RIGHT,
    allowed_areas: Sequence[DockArea | DockAreaString] | None = None,
    keybindings: KeyBindingsType | None = None,
    singleton: bool = False,
    plugin_configs: PluginConfigType | None = None,
    command_id: str | None = None,
    icon: str | None = None,
) -> _F: ...


@overload
def register_dock_widget_action(
    widget_factory: None = None,
    *,
    menus: str | Sequence[str] | None = None,
    title: str | None = None,
    area: DockArea | DockAreaString = DockArea.RIGHT,
    allowed_areas: Sequence[DockArea | DockAreaString] | None = None,
    keybindings: KeyBindingsType | None = None,
    singleton: bool = False,
    plugin_configs: PluginConfigType | None = None,
    command_id: str | None = None,
    icon: str | None = None,
) -> Callable[[_F], _F]: ...


def register_dock_widget_action(
    widget_factory=None,
    *,
    menus: str | Sequence[str] | None = None,
    title: str | None = None,
    area: DockArea | DockAreaString = DockArea.RIGHT,
    allowed_areas: Sequence[DockArea | DockAreaString] | None = None,
    keybindings=None,
    singleton: bool = False,
    plugin_configs: PluginConfigType | None = None,
    command_id: str | None = None,
    icon: str | None = None,
):
    """Register a widget factory as a dock widget function.

    Parameters
    ----------
    widget_factory : callable, optional
        Class of dock widget, or a factory function for the dock widget.
    menus : str or sequence of str, optional
        Menu ID or list of menu IDs where the action will be added.
    title : str, optional
        Title of the dock widget.
    area : DockArea or DockAreaString, optional
        Initial area of the dock widget.
    allowed_areas : sequence of DockArea or DockAreaString, optional
        List of areas that is allowed for the dock widget.
    keybindings : sequence of keybinding rule, optional
        Keybindings to trigger the dock widget.
    singleton : bool, default False
        If true, the registered dock widget will constructed only once.
    plugin_configs : dict, dataclass or pydantic.BaseModel, optional
        Default configuration for the plugin. This config will be saved in the
        application profile and will be used to update the dock widget via the method
        `update_configs(self, cfg) -> None`. This argument must be a dict, dataclass
        or pydantic.BaseModel. If a dict, the format must be like:

        ``` python
        plugin_configs = {
           "config_0": {"value": 0, "tooltip": ...},
           "config_1": {"value": "xyz", "tooltip": ...},
        }
        ```

        where only "value" is required. If a dataclass or pydantic.BaseModel, field
        objects will be used instead of the dict.

        ``` python
        @dataclass
        class MyPluginConfig:
            config_0: int = Field(default=0, metadata={"tooltip": ...})
            config_1: str = Field(default="xyz", metadata={"tooltip": ...})
        plugin_configs = MyPluginConfig()
        ```
    command_id : str, optional
        Command ID. If not given, the function name will be used.
    icon : str, optional
        Iconify icon key.
    """
    kbs = normalize_keybindings(keybindings)
    if menus is None:
        menus = [MenuId.TOOLS_DOCK]

    def _inner(wf: Callable):
        _command_id = command_id_from_func(wf, command_id)
        _callback = DockWidgetCallback(
            wf,
            title=title,
            area=area,
            allowed_areas=allowed_areas,
            singleton=singleton,
            uuid=uuid.uuid4(),
            command_id=_command_id,
        )
        if singleton:
            toggle_rule = ToggleRule(get_current=_callback.widget_visible)
        else:
            toggle_rule = None
        action = Action(
            id=_command_id,
            title=_callback._title,
            tooltip=tooltip_from_func(wf),
            callback=_callback,
            menus=norm_menus(menus),
            keybindings=kbs,
            toggled=toggle_rule,
            icon=icon,
            icon_visible_in_menu=False,
        )
        reg = AppActionRegistry.instance()
        reg.add_action(action)
        if plugin_configs:
            cfg_type = type(plugin_configs)
            reg._plugin_default_configs[_command_id] = PluginConfigTuple(
                _callback._title,
                plugin_configs,
                cfg_type,
            )
        return wf

    return _inner if widget_factory is None else _inner(widget_factory)


class WidgetCallbackBase(Generic[_W]):
    _instance_map = weakref.WeakValueDictionary[str, "Self"]()

    def __init__(
        self,
        func: Callable,
        title: str | None,
        uuid: uuid.UUID | None,
        command_id: str,
    ):
        self._func = func
        self._title = _normalize_title(title, func)
        self._uuid = uuid
        self._command_id = command_id
        # if singleton, retain the weak reference to the dock widget
        self._widget_ref: Callable[[], _W | None] = lambda: None
        self._all_widgets: weakref.WeakSet[_W] = weakref.WeakSet()
        wraps(func)(self)
        self.__annotations__ = {"ui": "MainWindow"}
        setattr(self, NO_RECORDING_FIELD, True)
        self.__class__._instance_map[command_id] = self

    @classmethod
    def instance_for_command_id(cls, command_id: str) -> Self | None:
        """Get the callback instance for the given command ID."""
        return WidgetCallbackBase._instance_map.get(command_id)


class DockWidgetCallback(WidgetCallbackBase["DockWidget"]):
    """Callback for registering dock widgets."""

    def __init__(
        self,
        func: Callable,
        title: str | None,
        area: DockArea | DockAreaString,
        allowed_areas: Sequence[DockArea | DockAreaString] | None,
        singleton: bool,
        uuid: uuid.UUID | None,
        command_id: str,
    ):
        super().__init__(func, title=title, uuid=uuid, command_id=command_id)
        self._singleton = singleton
        self._area = area
        self._allowed_areas = allowed_areas

    def __call__(self, ui: MainWindow) -> DockWidget:
        if self._singleton:
            if _dock := ui.dock_widgets.widget_for_id(self._uuid):
                _dock.visible = not _dock.visible
                return _dock
        try:
            widget = self._func(ui)
        except TypeError:
            widget = self._func()
        dock = ui.add_dock_widget(
            widget,
            title=self._title,
            area=self._area,
            allowed_areas=self._allowed_areas,
            _identifier=self._uuid,
        )
        dock._command_id = self._command_id
        self._all_widgets.add(dock)
        self._widget_ref = weakref.ref(dock)
        plugin_configs = ui.app_profile.plugin_configs.get(self._command_id)
        if plugin_configs:
            if not dock._has_update_configs:
                raise ValueError(
                    "The widget must have 'update_configs' method if plugin config "
                    "fields are given.",
                )
            params = {}
            for k, v in plugin_configs.items():
                params[k] = v["value"]
            dock.update_configs(params)
        return dock

    def widget_visible(self) -> bool:
        """Used for the toggle rule of the Action."""
        with suppress(RuntimeError):
            if widget := self._widget_ref():
                return widget.visible
        return False


def _normalize_title(title: str | None, func: Callable) -> str:
    if title is None:
        return func.__name__.replace("_", " ").title()
    return title


_C = TypeVar("_C", bound="PluginConfigType")


def _get_plugin_config(
    ui: MainWindow,
    config_class: type[_C],
    plugin_id: str | None = None,
) -> tuple[str, PluginConfigTuple]:
    if isinstance(config_class, str):
        if plugin_id is not None:
            raise TypeError("No overload matches the input.")
        _config_class, _plugin_id = None, config_class
    else:
        _config_class = config_class
        _plugin_id = plugin_id

    reg = AppActionRegistry.instance()
    if _plugin_id is None:
        for _id, _config in reg._plugin_default_configs.items():
            if isinstance(_config.config, _config_class):
                _plugin_id = _id
                break
        else:
            raise ValueError(
                f"Cannot find plugin ID for the config class: {config_class}."
            )
    config_dict = {
        k: v["value"] for k, v in ui.app_profile.plugin_configs[_plugin_id].items()
    }
    plugin_config = reg._plugin_default_configs[_plugin_id].updated(config_dict)
    if not isinstance(plugin_config.config, _config_class):
        raise TypeError(
            f"Plugin ID {_plugin_id} does not match the config class {config_class}."
        )
    return _plugin_id, plugin_config


def get_config(
    config_class: type[_C],
    plugin_id: str | None = None,
) -> _C | None:
    from himena.widgets import current_instance

    ui = current_instance()
    try:
        return _get_plugin_config(ui, config_class, plugin_id)[1].config
    except ValueError:
        return None


@overload
@contextmanager
def update_config_context(
    plugin_id: str,
    *,
    update_widget: bool = False,
) -> Iterator[PluginConfigType]: ...


@overload
@contextmanager
def update_config_context(
    config_class: type[_C],
    plugin_id: str | None = None,
    *,
    update_widget: bool = False,
) -> Iterator[_C]: ...


@contextmanager
def update_config_context(
    config_class: type | str,
    plugin_id: str | None = None,
    *,
    update_widget: bool = False,
):
    """Context manager for updating plugin config.

    If a config object is updated within the context, it will be saved to the profile
    and optionally trigger the update of the widget if `update_widget` is True.

    Examples
    --------

    ``` python
    with update_config_context(config_class=MyConfig) as cfg:
        cfg.some_attr = ...
    ```

    """
    from himena.widgets import current_instance

    ui = current_instance()
    plugin_id, plugin_config = _get_plugin_config(ui, config_class, plugin_id)
    cur_config = plugin_config.config
    yield cur_config
    prof = ui.app_profile
    all_configs = prof.plugin_configs.copy()

    cfg_dict = all_configs[plugin_id] = plugin_config.as_dict()
    prof.with_plugin_configs(all_configs).save()

    # update existing dock widgets with the new config
    if update_widget and (cb := WidgetCallbackBase.instance_for_command_id(plugin_id)):
        params = {}
        for key, opt in cfg_dict.items():
            params[key] = opt["value"]
        for widget in cb._all_widgets:
            # the internal widget should always has the method "update_configs"
            widget.update_configs(params)



from __future__ import annotations
from concurrent.futures import Future
from contextlib import suppress
import inspect
from timeit import default_timer as timer
import logging
from typing import Callable, Hashable, Literal, TypeVar, TYPE_CHECKING, cast
from pathlib import Path
import warnings
import weakref
import numpy as np

from qtpy import QtWidgets as QtW, QtGui, QtCore
from app_model.types import KeyCode
from app_model.backends.qt import (
    QModelMainWindow,
    QModelMenu,
    QMenuItemAction,
    QModelToolBar,
    QCommandRuleAction,
    qkey2modelkey,
)
from superqt import QIconifyIcon
from superqt.utils import ensure_main_thread
from himena.consts import MenuId
from himena.consts import ParametricWidgetProtocolNames as PWPN
from himena.utils.window_rect import prevent_window_overlap
from himena._app_model import _formatter, HimenaApplication
from himena.plugins import AppActionRegistry
from himena._utils import doc_to_whats_this
from himena.qt._qnotification import QJobStack, QNotificationWidget, QWhatsThisWidget
from himena.qt._qtab_widget import QTabWidget
from himena.qt._qstatusbar import QStatusBar
from himena.qt._qsub_window import QSubWindow, QSubWindowArea, get_subwindow
from himena.qt._qdock_widget import QDockWidget
from himena.qt._qcommand_palette import QCommandPalette
from himena.qt._qcontrolstack import QControlStack
from himena.qt._qparametric import QParametricWidget
from himena.qt._qgoto import QGotoWidget
from himena.qt._qresult_stack import QResultStack
from himena.types import (
    DockArea,
    DockAreaString,
    ClipboardDataModel,
    Size,
    WindowState,
    WindowRect,
    BackendInstructions,
)
from himena.app import get_event_loop_handler
from himena import widgets
from himena.qt.registry import list_widget_class
from himena.qt._utils import get_stylesheet_path, ArrayQImage, ndarray_to_qimage

if TYPE_CHECKING:
    from himena.widgets._main_window import SubWindow, MainWindow
    from himena.style import Theme

_ICON_PATH = Path(__file__).parent.parent / "resources"
_T = TypeVar("_T", bound=QtW.QWidget)
_V = TypeVar("_V")
_LOGGER = logging.getLogger(__name__)


class QMainWindow(QModelMainWindow, widgets.BackendMainWindow[QtW.QWidget]):
    """The Qt mainwindow implementation for himena."""

    _himena_main_window: MainWindow
    _app: HimenaApplication
    status_tip_requested = QtCore.Signal(str, float)
    notification_requested = QtCore.Signal(str, float)

    def __init__(self, app: HimenaApplication):
        # app must be initialized
        QtCore.QDir.addSearchPath("himena-icons", _ICON_PATH.as_posix())
        _event_loop_handler = get_event_loop_handler(
            backend="qt",
            app_name=app.name,
            host=app.attributes.get("host", "localhost"),
            port=app.attributes.get("port", 49200),
        )
        self._qt_app = cast(QtW.QApplication, _event_loop_handler.get_app())
        self._app_name = app.name
        self._keys_down: set[int] = set()
        self._event_loop_handler = _event_loop_handler

        super().__init__(app)
        self._qt_app.setApplicationName(app.name)
        self.setWindowTitle("himena")
        self.setWindowIcon(QtGui.QIcon(_ICON_PATH.joinpath("icon.svg").as_posix()))
        self._tab_widget = QTabWidget()
        self._menubar = self.setModelMenuBar(_prep_menubar_map(app))
        self._menubar.setContextMenuPolicy(
            QtCore.Qt.ContextMenuPolicy.PreventContextMenu
        )

        # Toolbar
        self._toolbar = self.addModelToolBar(menu_id=MenuId.TOOLBAR)
        _init_tool_bar(self._toolbar)
        self._toolbar.setFixedHeight(32)
        self._toolbar.addSeparator()
        self._control_stack = QControlStack(self._toolbar)
        self._toolbar.addWidget(self._control_stack)

        self.setCentralWidget(self._tab_widget)

        self._status_bar = QStatusBar(self)
        self.setStatusBar(self._status_bar)

        self._command_palette_general = QCommandPalette(
            self._app,
            parent=self,
            formatter=_formatter.formatter_general,
        )
        self._command_palette_recent = QCommandPalette(
            self._app,
            menu_id=MenuId.RECENT_ALL,
            parent=self,
            exclude=["open-recent"],
            formatter=_formatter.formatter_recent,
            placeholder="Open recent file ...",
        )
        self._command_palette_new = QCommandPalette(
            self._app,
            menu_id=MenuId.FILE_NEW,
            parent=self,
            exclude=["new"],
            formatter=_formatter.formatter_general,
            placeholder="New ...",
        )
        self._goto_widget = QGotoWidget(self)

        self._anim_subwindow = QtCore.QPropertyAnimation()
        self._confirm_close = True
        self.setMinimumSize(400, 300)
        self.resize(1000, 720)

        # connect notifications
        self._event_loop_handler.errored.connect(self._on_error)
        self._event_loop_handler.warned.connect(self._on_warning)
        self.status_tip_requested.connect(self._on_status_tip_requested)
        self.notification_requested.connect(self._on_show_notification_requested)

        # job stack
        self._job_stack = QJobStack(self)
        self._job_stack.setAnchor("bottom_left")

        # top right buttons to menu
        self._corner_toolbar = QCornerToolBar(MenuId.CORNER, app, parent=self)
        _init_tool_bar(self._corner_toolbar)
        self._corner_toolbar.setIconSize(QtCore.QSize(16, 16))
        self._corner_toolbar.setContentsMargins(0, 0, 0, 0)
        self._menubar.setCornerWidget(self._corner_toolbar)

    def _update_widget_theme(self, style: Theme):
        self.setStyleSheet(style.format_text(get_stylesheet_path().read_text()))
        if style.is_light_background():
            icon_color = "#333333"
        else:
            icon_color = "#CCCCCC"
        _update_toolbtn_color(self._toolbar, icon_color)
        _update_toolbtn_color(self._corner_toolbar, icon_color)
        for i in range(self._tab_widget.count()):
            if area := self._tab_widget.widget_area(i):
                for sub in area.subWindowList():
                    sub._title_bar._set_icon_color(icon_color)

    def _main_window_rect(self) -> WindowRect:
        geo = self.geometry()
        return WindowRect(geo.x(), geo.y(), geo.width(), geo.height())

    def _set_main_window_rect(self, rect: WindowRect) -> None:
        self.setGeometry(rect.left, rect.top, rect.width, rect.height)

    @ensure_main_thread
    def _on_error(self, exc: Exception) -> None:
        from himena.qt._qtraceback import QtErrorMessageBox

        mbox = QtErrorMessageBox.from_exc(exc, parent=self)
        notification = QNotificationWidget(self)
        notification.addWidget(mbox)
        return notification.show_and_hide_later()

    @ensure_main_thread
    def _on_warning(self, warning: warnings.WarningMessage) -> None:
        from himena.qt._qtraceback import QtErrorMessageBox

        mbox = QtErrorMessageBox.from_warning(warning, parent=self)
        notification = QNotificationWidget(self)
        notification.addWidget(mbox)
        return notification.show_and_hide_later()

    def add_dock_widget(
        self,
        widget: QtW.QWidget,
        *,
        title: str | None = None,
        area: DockAreaString | DockArea | None = DockArea.RIGHT,
        allowed_areas: list[DockAreaString | DockArea] | None = None,
    ) -> QDockWidget:
        # Normalize title and areas
        if title is None:
            title = widget.objectName()
        # Construct and add the dock widget
        dock_widget = QDockWidget(widget, title, allowed_areas)
        self.addDockWidget(dock_widget.area_normed(area), dock_widget)
        dock_widget.closed.connect(self._update_context)
        if doc := getattr(widget, "__doc__", ""):
            dock_widget.whats_this.connect(lambda: self._show_dock_whats_this(doc))
        QtW.QApplication.processEvents()
        return dock_widget

    def _dock_widget_title(self, widget: QtW.QWidget) -> str:
        if isinstance(dock := widget.parentWidget(), QtW.QDockWidget):
            return dock.windowTitle()
        raise ValueError(f"{widget!r} does not have a dock widget parent.")

    def _set_dock_widget_title(self, widget: QtW.QWidget, title: str) -> None:
        with suppress(RuntimeError):
            if isinstance(dock := widget.parentWidget(), QtW.QDockWidget):
                return dock.setWindowTitle(title)

    def _dock_widget_visible(self, widget: QtW.QWidget) -> bool:
        with suppress(RuntimeError):
            # RuntimeError: wrapped C/C++ object of type QDockWidget has been deleted
            if isinstance(dock := widget.parentWidget(), QtW.QDockWidget):
                return dock.isVisible()
        return False

    def _set_dock_widget_visible(self, widget: QtW.QWidget, visible: bool) -> None:
        with suppress(RuntimeError):
            if isinstance(dock := widget.parentWidget(), QtW.QDockWidget):
                return dock.setVisible(visible)

    def _set_control_widget(self, widget: QtW.QWidget, control: QtW.QWidget) -> None:
        if not isinstance(control, QtW.QWidget):
            warnings.warn(
                f"`control_widget()` of widget class {type(widget)} did not return a "
                f"QWidget (got {type(control)}).",
                UserWarning,
                stacklevel=2,
            )
        else:
            self._control_stack.add_control_widget(widget, control)

    def _update_control_widget(self, current: QtW.QWidget) -> None:
        self._control_stack.update_control_widget(current)

    def _remove_control_widget(self, widget: QtW.QWidget) -> None:
        self._control_stack.remove_control_widget(widget)

    def _del_dock_widget(self, widget: QtW.QWidget) -> None:
        if isinstance(dock := widget.parentWidget(), QtW.QDockWidget):
            dock.close()

    def add_tab(self, tab_name: str) -> QSubWindowArea:
        """Add a new tab with a sub-window area.

        Parameters
        ----------
        tab_name : str
            Name of the tab.
        """
        _LOGGER.info("Adding tab of title %r", tab_name)
        return self._tab_widget.add_tab_area(tab_name)

    def add_widget(
        self,
        widget: _T,
        i_tab: int,
        title: str,
    ) -> QSubWindow:
        if not isinstance(widget, QtW.QWidget):
            raise TypeError(
                f"Expected a QWidget, got {type(widget)}. If you are adding a wrapper "
                "of a widget, the wrapper should implement `native_widget()` method "
                "that returns QWidget."
            )
        tab = self._tab_widget.widget_area(i_tab)
        _LOGGER.info("Adding widget of title %r to tab %r", title, i_tab)
        subwindow = tab.add_widget(widget, title)
        if self._himena_main_window.theme.is_light_background():
            icon_color = "#333333"
        else:
            icon_color = "#CCCCCC"
        subwindow._set_icon_color(icon_color)
        return subwindow

    def set_widget_as_preview(self, subwindow: SubWindow):
        qsubwindow = get_subwindow(subwindow._split_interface_and_frontend()[1])
        qsubwindow._title_bar._model_menu_btn.hide()
        qsubwindow._title_bar._window_menu_btn.hide()

    def _connect_window_events(self, sub: SubWindow, qsub: QSubWindow):
        @qsub.state_change_requested.connect
        def _(state: WindowState):
            sub._set_state(state)

        @qsub.close_requested.connect
        def _():
            main = self._himena_main_window
            sub._close_me(main, main._instructions.confirm)

        @qsub.rename_requested.connect
        def _(title: str):
            sub.title = title

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        super().showEvent(event)
        self._himena_main_window._main_window_resized(Size(*self._area_size()))
        if self.statusTip() == "":
            self._try_show_default_status_tip()

    def show(self, run: bool = False):
        super().show()
        self.activateWindow()
        size = self.size()
        minw, minh = 600, 400
        if size.width() < minw or size.height() < minh:
            self.resize(min(size.width(), minw), min(size.height(), minh))
        if run:
            get_event_loop_handler("qt", self._app_name).run_app()

    def _try_show_default_status_tip(self) -> None:
        if tip := AppActionRegistry.instance().pick_a_tip():
            self._set_status_tip(f"Tip: {tip.short} │ {tip.long}", 8.0)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        if event.spontaneous() and self._confirm_close and not self._ok_to_exit():
            event.ignore()
            return
        if info := self._status_bar._profile_info:
            info.close()
        self._event_loop_handler.close_socket()
        return super().closeEvent(event)

    def focusOutEvent(self, a0):
        if area := self._tab_widget.current_widget_area():
            area._tooltip_widget.hide()
        return super().focusOutEvent(a0)

    def event(self, e: QtCore.QEvent) -> bool:
        if e.type() in {
            QtCore.QEvent.Type.WindowActivate,
            QtCore.QEvent.Type.ZOrderChange,
        }:
            # upon activation or raise_, put window at the end of _instances
            widgets.set_current_instance(self._app_name, self._himena_main_window)

        res = super().event(e)

        if e.type() == QtCore.QEvent.Type.Close and e.isAccepted():
            widgets.remove_instance(self._app_name, self._himena_main_window)

        return res

    def _ok_to_exit(self) -> bool:
        ui = self._himena_main_window
        if any(win._need_ask_save_before_close() for win in ui.iter_windows()):
            res = ui.exec_choose_one_dialog(
                title="Closing",
                message="There are unsaved changes. Exit anyway?",
                choices=["Exit", "Cancel"],
            )
            return res == "Exit"
        return True

    def _update_context(self) -> None:
        _time_0 = timer()
        ctx = self._himena_main_window._ctx_keys
        ctx._update(self._himena_main_window)
        _dict = ctx.dict()
        self._menubar.update_from_context(_dict)
        self._toolbar.update_from_context(_dict)
        self._corner_toolbar.update_from_context(_dict)
        _msec = (timer() - _time_0) * 1000
        _LOGGER.debug("Context update took %.3f msec", _msec)

    def _current_tab_index(self) -> int | None:
        idx = self._tab_widget.currentIndex()
        if idx < 0:
            return None
        return idx

    def _set_current_tab_index(self, i_tab: int) -> None:
        return self._tab_widget.setCurrentIndex(i_tab)

    def _tab_hash(self, i_tab: int) -> Hashable:
        return self._tab_widget.widget_area(i_tab)

    def _tab_hash_for_window(self, widget: QtW.QWidget) -> Hashable:
        mdiarea = get_subwindow(widget)._qt_mdiarea()
        assert isinstance(mdiarea, QSubWindowArea)
        return mdiarea

    def _num_tabs(self) -> int:
        if self._tab_widget._is_startup_only():
            return 0
        return self._tab_widget.count()

    def _current_sub_window_index(self, i_tab: int) -> int | None:
        area = self._tab_widget.widget_area(i_tab)
        if area is None:
            return None
        sub = area.currentSubWindow()
        if sub is None or not sub.is_current():
            return None
        return area.subWindowList().index(sub)

    def _set_current_sub_window_index(self, i_tab: int, i_window: int) -> None:
        assert i_window is None or i_window >= 0
        area = self._tab_widget.widget_area(i_tab)
        subwindows = area.subWindowList()
        for i in range(len(subwindows)):
            subwindows[i].set_is_current(i == i_window)
        if i_window is not None:
            area.setActiveSubWindow(subwindows[i_window])

    def _tab_title(self, i_tab: int) -> str:
        return self._tab_widget.tabText(i_tab)

    def _set_tab_title(self, i_tab: int, title: str) -> None:
        return self._tab_widget.setTabText(i_tab, title)

    def _window_title(self, widget: QtW.QWidget) -> str:
        window = get_subwindow(widget)
        return window.windowTitle()

    def _set_window_title(self, widget: QtW.QWidget, title: str) -> None:
        window = get_subwindow(widget)
        return window.setWindowTitle(title)

    def _list_widget_class(self, type: str):
        return list_widget_class(self._app_name, type)

    def _open_file_dialog(
        self,
        mode: str = "r",
        extension_default: str | None = None,
        allowed_extensions: list[str] | None = None,
        caption: str | None = None,
        start_path: Path | None = None,
    ) -> Path | list[Path] | None:
        if allowed_extensions:
            filter_str = (
                ";".join(_ext_to_filter(ext) for ext in allowed_extensions)
                + ";;All Files (*)"
            )
        else:
            filter_str = "All Files (*)"

        match mode:
            case "r":
                path, _ = QtW.QFileDialog.getOpenFileName(
                    self,
                    caption=caption or "Open File",
                    directory=start_path.as_posix() if start_path else None,
                    filter=filter_str,
                )
                if path:
                    return Path(path)
            case "w":
                path, _ = QtW.QFileDialog.getSaveFileName(
                    self,
                    caption=caption or "Save File",
                    directory=start_path.as_posix() if start_path else None,
                    filter=filter_str,
                )
                if path:
                    output_path = Path(path)
                    if output_path.suffix == "" and extension_default is not None:
                        output_path = output_path.with_suffix(extension_default)
                    return output_path
            case "rm":
                paths, _ = QtW.QFileDialog.getOpenFileNames(
                    self,
                    caption=caption or "Open Files",
                    directory=start_path.as_posix() if start_path else None,
                    filter=filter_str,
                )
                if paths:
                    return [Path(p) for p in paths]
            case "d":
                path = QtW.QFileDialog.getExistingDirectory(
                    self,
                    caption=caption or "Open Directory",
                    directory=start_path.as_posix() if start_path else None,
                )
                if path:
                    return Path(path)
            case _:
                raise ValueError(f"Invalid file mode {mode!r}.")

    def _request_choice_dialog(
        self,
        title: str,
        message: str,
        choices: list[tuple[str, _V]],
        how: Literal["buttons", "radiobuttons"] = "buttons",
    ) -> _V | None:
        match how:
            case "buttons":
                return QChoicesDialog.request(title, message, choices, parent=self)
            case "radiobuttons":
                return QChoicesDialog.request_radiobuttons(
                    title, message, choices, parent=self
                )
            case _:
                raise NotImplementedError

    def _show_command_palette(
        self,
        kind: Literal["general", "recent", "goto", "new"],
    ) -> None:
        match kind:
            case "general":
                self._command_palette_general.update_context(self)
                self._command_palette_general.show()
            case "recent":
                self._command_palette_general.update_context(self)
                self._command_palette_recent.show()
            case "goto":
                self._goto_widget.show()
                self._goto_widget.setFocus()
            case "new":
                self._command_palette_new.update_context(self)
                self._command_palette_new.show()
            case _:
                raise NotImplementedError

    def _exit_main_window(self, confirm: bool = False) -> None:
        if confirm and not self._ok_to_exit():
            return None
        return self.close()

    def _get_tab_name_list(self) -> list[str]:
        if self._tab_widget._is_startup_only():
            return []
        return [self._tab_widget.tabText(i) for i in range(self._tab_widget.count())]

    def _get_widget_list(self, i_tab: int) -> list[tuple[str, QtW.QWidget]]:
        tab = self._tab_widget.widget_area(i_tab)
        if tab is None:
            return []
        return [(w.windowTitle(), w._widget) for w in tab.subWindowList()]

    def _del_widget_at(self, i_tab: int, i_window: int) -> None:
        if i_tab < 0 or i_window < 0:
            raise ValueError("Invalid tab or window index.")
        _LOGGER.info("Deleting widget at tab %r, window %r", i_tab, i_window)
        tab = self._tab_widget.widget_area(i_tab)
        subwindows = tab.subWindowList()
        subwindow = subwindows[i_window]
        tab.removeSubWindow(subwindow)
        subwindow._widget._himena_widget = None
        tab.relabel_widgets()
        if len(subwindows) == 1:  # now empty
            tab._set_area_focused()

    def _del_tab_at(self, i_tab: int) -> None:
        _LOGGER.info("Deleting tab at index %r", i_tab)
        self._tab_widget.remove_tab_area(i_tab)
        if self._tab_widget._is_startup_only():
            self._try_show_default_status_tip()
        QtW.QApplication.processEvents()

    def _rename_window_at(self, i_tab: int, i_window: int) -> None:
        tab = self._tab_widget.widget_area(i_tab)
        window = tab.subWindowList()[i_window]
        window._title_bar._start_renaming()

    def _window_state(self, widget: QtW.QWidget) -> WindowState:
        return get_subwindow(widget).state

    def _set_window_state(
        self,
        widget: QtW.QWidget,
        state: WindowState,
        inst: BackendInstructions,
    ) -> None:
        get_subwindow(widget)._update_window_state(state, animate=inst.animate)

    def _window_rect(self, widget: QtW.QWidget) -> WindowRect:
        geo = get_subwindow(widget).geometry()
        return WindowRect(geo.x(), geo.y(), geo.width(), geo.height())

    def _set_window_rect(
        self,
        widget: QtW.QWidget,
        rect: WindowRect,
        inst: BackendInstructions,
    ) -> None:
        qrect = QtCore.QRect(rect.left, rect.top, rect.width, rect.height)
        if inst.animate:
            get_subwindow(widget)._set_geometry_animated(qrect)
        else:
            get_subwindow(widget).setGeometry(qrect)

    def _area_size(self) -> tuple[int, int]:
        if widget := self._tab_widget.currentWidget():
            size = widget.size()
        else:
            size = self._tab_widget.children()[0].size()
        return size.width(), size.height()

    def _clipboard_data(self) -> ClipboardDataModel | None:
        clipboard = QtGui.QGuiApplication.clipboard()
        if clipboard is None:
            return None
        md = clipboard.mimeData()
        model = ClipboardDataModel()
        if md is None:
            return None
        if md.hasHtml():
            model.html = md.html()
        if md.hasImage():
            model.image = ArrayQImage(clipboard.image())
        if md.hasText():
            model.text = md.text()
        if md.hasUrls():
            model.files = [Path(url.toLocalFile()) for url in md.urls()]
        return model

    @ensure_main_thread
    def _set_clipboard_data(self, data: ClipboardDataModel) -> None:
        clipboard = QtW.QApplication.clipboard()
        if clipboard is None:
            return
        mime = QtCore.QMimeData()
        if (html := data.html) is not None:
            mime.setHtml(html)
        if (text := data.text) is not None:
            mime.setText(text)
        if (img := data.image) is not None:
            if not isinstance(img, ArrayQImage):
                qimg = ndarray_to_qimage(np.asarray(img, dtype=np.uint8))
            else:
                qimg = img.qimage
            mime.setImageData(qimg)
        if (files := data.files) is not None:
            mime.setUrls([QtCore.QUrl.fromLocalFile(str(f)) for f in files])
        return clipboard.setMimeData(mime)

    def _connect_main_window_signals(self, main: MainWindow):
        self._tab_widget.currentChanged.connect(main._tab_activated)
        self._tab_widget.activeWindowChanged.connect(main._window_activated)
        self._tab_widget.resized.connect(self._emit_resized)

    def _emit_resized(self):
        self._himena_main_window._main_window_resized(Size(*self._area_size()))

    def _screenshot(self, target: str):
        match target:
            case "main":
                qwidget = self
            case "area":
                if widget := self._tab_widget.current_widget_area():
                    qwidget = widget
                else:
                    raise ValueError("No active area.")
            case "window":
                if sub := self._tab_widget.current_widget_area().currentSubWindow():
                    qwidget = sub._widget
                else:
                    raise ValueError("No active window.")
            case tgt:
                raise ValueError(f"Invalid target name {tgt!r}.")
        return ArrayQImage.from_qwidget(qwidget)

    def _process_parametric_widget(
        self,
        widget: QtW.QWidget,
    ) -> QtW.QWidget:
        return QParametricWidget(widget)

    def _connect_parametric_widget_events(
        self,
        wrapper: widgets.ParametricWindow[QParametricWidget],
        widget: QParametricWidget,
    ) -> None:
        widget._call_btn.clicked.connect(wrapper._emit_btn_clicked)
        widget.param_changed.connect(wrapper._emit_param_changed)

    def _signature_to_widget(
        self,
        sig: inspect.Signature,
        show_parameter_labels: bool = True,
        preview: bool = False,
    ) -> QtW.QWidget:
        from magicgui.signature import MagicSignature
        from himena.qt.magicgui import ToggleSwitch, get_type_map

        container = MagicSignature.from_signature(sig).to_container(
            type_map=get_type_map(),
            labels=show_parameter_labels,
        )
        container.margins = (0, 0, 0, 0)
        setattr(container, PWPN.GET_PARAMS, container.asdict)
        setattr(container, PWPN.UPDATE_PARAMS, container.update)
        setattr(container, PWPN.CONNECT_CHANGED_SIGNAL, container.changed.connect)
        if preview:
            checkbox = ToggleSwitch(value=False, text="Preview")
            container.append(checkbox)
            setattr(container, PWPN.IS_PREVIEW_ENABLED, checkbox.get_value)
        return container

    def _add_widget_to_parametric_window(
        self,
        wrapper: widgets.ParametricWindow[QParametricWidget],
        widget: QtW.QWidget,
        result_as: Literal["below", "right"],
    ):
        match result_as:
            case "below":
                wrapper.widget.add_widget_below(widget)
            case "right":
                wrapper.widget.add_widget_right(widget)
            case _:
                raise ValueError(f"Invalid result_as value {result_as!r}.")
        if wrapper.is_preview_enabled() and not wrapper._auto_close:
            wrapper.widget._call_btn.hide()

    def _remove_widget_from_parametric_window(
        self,
        wrapper: widgets.ParametricWindow[QParametricWidget],
    ):
        wrapper.widget.remove_result_widget()

    def _move_focus_to(self, win: QtW.QWidget) -> None:
        win.setFocus()

    def _set_status_tip(self, tip: str, duration: float) -> None:
        self.status_tip_requested.emit(tip, duration)

    def _show_notification(self, text: str, duration: float) -> None:
        self.notification_requested.emit(text, duration)

    def _show_tooltip(self, text: str, duration: float, behavior: str) -> None:
        if area := self._tab_widget.current_widget_area():
            area._tooltip_widget.set_behavior(behavior)
            area._tooltip_widget.show_tooltip(text, duration)

    def _on_status_tip_requested(self, tip: str, duration: float) -> None:
        self._status_bar.showMessage(tip, int(duration * 1000))

    def _on_show_notification_requested(self, text: str, duration: float) -> None:
        text_edit = QtW.QPlainTextEdit(text)
        text_edit.setWordWrapMode(QtGui.QTextOption.WrapMode.WordWrap)
        notification = QNotificationWidget(self)
        notification.addWidget(text_edit)
        notification.show_and_hide_later(duration)

    def _get_menu_action_by_id(self, name: str) -> QtW.QAction:
        # Find the help menu
        for action in self._menubar.actions():
            if isinstance(menu := action.menu(), QModelMenu):
                if menu._menu_id == name:
                    action = action
                    break
        else:
            raise RuntimeError(f"{name} menu not found.")
        return action

    def _rebuild_for_runtime(self, added_menus: list[str]) -> None:
        # Find the help menu
        help_menu_action = self._get_menu_action_by_id("help")

        # insert the new menus to the menubar before the help menu
        for menu_id in added_menus:
            menu = QModelMenu(menu_id, self._app, menu_id.title(), self._menubar)
            self._menubar.insertMenu(help_menu_action, menu)
        # rebuild the menus
        for action in self._menubar.actions():
            if isinstance(menu := action.menu(), QModelMenu):
                menu.rebuild()

    def _process_future_done_callback(
        self,
        cb: Callable[[Future], None],
        cb_errored: Callable[[Exception], None],
        **kwargs,
    ) -> Callable[[Future], None]:
        def _func(future: Future):
            if future.cancelled():
                pass
            elif e := future.exception():
                self._on_error(e)
                cb_errored(e)
            else:
                cb(future, **kwargs)

        return ensure_main_thread(_func)

    def _set_parametric_widget_busy(
        self,
        wrapper: widgets.ParametricWindow[QParametricWidget],
        busy: bool,
    ):
        """Set the parametric widget busy status."""
        wrapper.widget.set_busy(busy)

    def _add_job_progress(self, future: Future, desc: str, total: int = 0) -> None:
        self._job_stack.add_future(future, desc, total)

    def _add_whats_this(
        self,
        text: str,
        style: Literal["plain", "markdown", "html"] = "plain",
    ) -> None:
        whatsthis = QWhatsThisWidget(self)
        whatsthis.set_text(text, style)
        whatsthis.show()

    def _show_dock_whats_this(self, doc: str):
        doc_formatted = doc_to_whats_this(doc)
        self._add_whats_this(doc_formatted, style="markdown")

    @ensure_main_thread
    def _append_result(self, item):
        """Add item to the result stack."""
        ui = self._himena_main_window
        tab = ui.tabs.current()
        if tab is None:
            tab = ui.add_tab()
        if (result_stack := tab._result_stack_ref()) is None:
            result_stack = QResultStack()
            tab._result_stack_ref = weakref.ref(result_stack)
            sub = tab.current()
            win = tab.add_widget(result_stack, title="Results")
            win.closed.connect(tab._discard_result_stack_ref)
            if sub:
                win.rect = prevent_window_overlap(sub, win, self._area_size())
        result_stack.append_result(item)

    def _keys_as_set(self) -> set[int]:
        out = set()
        for k in self._keys_down:
            model_key = qkey2modelkey(k)
            if isinstance(model_key, KeyCode):
                out.add(int(model_key))
            else:
                out.add(int(model_key._key))
        return out


def _is_root_menu_id(app: HimenaApplication, menu_id: str) -> bool:
    if menu_id in (MenuId.TOOLBAR, MenuId.CORNER, app.menus.COMMAND_PALETTE_ID):
        return False
    if len(menu_id) == 0:
        return False
    return "/" not in menu_id.replace("//", "")


def _ext_to_filter(ext: str) -> str:
    if ext.startswith("."):
        return f"*{ext}"
    elif ext == "":
        return "*"
    else:
        return f"*.{ext}"


class QCornerToolBar(QModelToolBar):
    # Managed by MenuId.CORNER
    def addSeparator(self):
        """No separator."""

    def update_from_context(self, ctx):
        super().update_from_context(ctx)
        # immediately update the check state of all actions
        for action in self.actions():
            if isinstance(action, QCommandRuleAction):
                action._refresh()


class QChoicesDialog(QtW.QDialog):
    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self._result = None
        self.accepted.connect(self.close)
        self._layout = QtW.QVBoxLayout(self)
        self._layout.setSpacing(10)

    def set_result_callback(self, value: _V, accept: bool) -> None:
        def _set_result():
            self._result = value
            if accept:
                return self.accept()

        return _set_result

    @classmethod
    def make_request(
        cls,
        title: str,
        message: str,
        choices: list[tuple[str, _V]],
        parent: QtW.QWidget | None = None,
    ) -> QChoicesDialog:
        self = cls(parent)
        self.init_message(title, message)
        button_group = QtW.QDialogButtonBox(self)
        shortcut_registered = set()
        for choice, value in choices:
            button = QtW.QPushButton(choice)
            button_group.addButton(button, button_group.ButtonRole.AcceptRole)
            button.clicked.connect(self.set_result_callback(value, accept=True))
            shortcut = choice[0].lower()
            if shortcut not in shortcut_registered:
                button.setShortcut(shortcut)
                shortcut_registered.add(shortcut)
        self._layout.addWidget(button_group)
        return self

    @classmethod
    def request(
        cls,
        title: str,
        message: str,
        choices: list[tuple[str, _V]],
        parent: QtW.QWidget | None = None,
    ) -> _V | None:
        self = cls.make_request(title, message, choices, parent)
        if self.exec() == QtW.QDialog.DialogCode.Accepted:
            return self._result

    @classmethod
    def make_request_radiobuttons(
        cls,
        title: str,
        message: str,
        choices: list[tuple[str, _V]],
        parent: QtW.QWidget | None = None,
    ) -> QChoicesDialog:
        self = cls(parent)
        self.init_message(title, message)
        button_group = QtW.QButtonGroup(self)
        button_group.setExclusive(True)
        for choice, value in choices:
            button = QtW.QRadioButton(choice)
            button_group.addButton(button)
            button.clicked.connect(self.set_result_callback(value, accept=False))
            self._layout.addWidget(button)

        ok_cancel = QtW.QDialogButtonBox()
        ok_cancel.addButton(QtW.QDialogButtonBox.StandardButton.Ok)
        ok_cancel.addButton(QtW.QDialogButtonBox.StandardButton.Cancel)
        ok_cancel.accepted.connect(self.accept)
        ok_cancel.rejected.connect(self.reject)
        self._layout.addWidget(ok_cancel)
        return self

    @classmethod
    def request_radiobuttons(
        cls,
        title: str,
        message: str,
        choices: list[tuple[str, _V]],
        parent: QtW.QWidget | None = None,
    ) -> _V | None:
        self = cls.make_request_radiobuttons(title, message, choices, parent)
        if self.exec() == QtW.QDialog.DialogCode.Accepted:
            return self._result

    def init_message(self, title: str, message: str) -> None:
        self.setWindowTitle(title)
        label = QtW.QLabel(message)
        self._layout.addWidget(label)


def _prep_menubar_map(app: HimenaApplication) -> dict[str, str]:
    default_menu_ids = {
        MenuId.FILE: "&" + MenuId.FILE.capitalize(),
        MenuId.WINDOW: "&" + MenuId.WINDOW.capitalize(),
        MenuId.VIEW: "&" + MenuId.VIEW.capitalize(),
        MenuId.TOOLS: "&" + MenuId.TOOLS.capitalize(),
        MenuId.GO: "&" + MenuId.GO.capitalize(),
    }
    existing_chars = {"f", "w", "v", "t", "g"}
    for menu_id, _ in app.menus:
        if menu_id not in default_menu_ids and _is_root_menu_id(app, menu_id):
            if menu_id and menu_id[0].lower() in existing_chars:
                title = menu_id.replace("_", " ").title()
            else:
                title = "&" + menu_id.replace("_", " ").title()
            default_menu_ids[menu_id] = title
    return default_menu_ids


def _init_tool_bar(tbar: QModelToolBar):
    tbar.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.PreventContextMenu)
    tbar.setMovable(False)


def _update_toolbtn_color(toolbar: QModelToolBar, icon_color: str):
    assert isinstance(toolbar._app, HimenaApplication)
    for action in toolbar.actions():
        if isinstance(action, QMenuItemAction):
            btn = toolbar.widgetForAction(action)
            if not isinstance(btn, QtW.QToolButton):
                continue
            icon = toolbar._app.registered_actions[action._command_id].icon
            if icon is not None:
                qicon = QIconifyIcon(icon.light, color=icon_color)
                btn.setIcon(qicon)
                btn.actions()[0].setIcon(qicon)



from himena.data_wrappers import wrap_array
import numpy as np
import xarray as xr
import pytest

rng = np.random.default_rng(1234)

@pytest.mark.parametrize(
    "arr",
    [
        rng.integers(0, 255, size=(2, 3, 4), dtype=np.uint8),
        xr.DataArray(rng.integers(0, 255, size=(2, 3, 4), dtype=np.uint8)),
    ]
)
def test_arrays(arr):
    ar = wrap_array(arr)
    assert ar.dtype == np.uint8
    assert ar.ndim == 3
    assert ar.shape == (2, 3, 4)
    assert ar.size == 24



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy import QtCore
from himena._descriptors import SaveToPath, CannotSave
from himena._providers import ReaderStore

if TYPE_CHECKING:
    from himena.widgets import SubWindow


class QWatchFileObject(QtCore.QObject):
    _instances = set()

    def __init__(self, win: SubWindow):
        super().__init__()
        if not isinstance(sb := win.save_behavior, SaveToPath):
            raise ValueError("no file to watch")
        if not win.supports_update_model:
            raise ValueError(f"Window {win.title!r} does not implement `update_model`.")
        self._subwindow = win
        self._old_save_behavior = sb
        self._file_path = sb.path
        self._watcher = QtCore.QFileSystemWatcher([str(sb.path)])

        win.closed.connect(self._on_target_window_closed)
        self._watcher.fileChanged.connect(self._on_file_change)
        win._save_behavior = CannotSave(reason="File watching is enabled")
        self.__class__._instances.add(self)

    def _on_file_change(self):
        ins = ReaderStore.instance()
        model = ins.run(self._file_path, plugin=self._old_save_behavior.plugin)
        if self._subwindow.is_alive:
            self._subwindow.update_model(model)

    def _on_target_window_closed(self):
        self._watcher.removePaths([str(self._file_path)])
        self._watcher.fileChanged.disconnect(self._on_file_change)
        self._instances.discard(self)
        self._subwindow._save_behavior = self._old_save_behavior
        self._subwindow.closed.disconnect(self._on_target_window_closed)



from himena import StandardType
from himena.plugins import register_widget_class
from himena_builtins.qt.widgets.table import QSpreadsheet, SpreadsheetConfigs

register_widget_class(
    StandardType.TABLE, QSpreadsheet, plugin_configs=SpreadsheetConfigs()
)



from __future__ import annotations

from pathlib import Path
from logging import getLogger
from typing import Generic, Iterator, TypeVar, TYPE_CHECKING
import warnings
from himena.utils.misc import PluginInfo
from himena.types import WidgetDataModel

if TYPE_CHECKING:
    from typing import Self
    from himena.plugins.io import ReaderPlugin, WriterPlugin, _IOPluginBase

    PathOrPaths = str | Path | list[str | Path]

_LOGGER = getLogger(__name__)
_S = TypeVar("_S", bound="_IOPluginBase")


class PluginStore(Generic[_S]):
    _global_instance = None

    def __init__(self):
        self._plugin_items: list[_S] = []

    @classmethod
    def instance(cls) -> Self:
        if cls._global_instance is None:
            cls._global_instance = cls()
        return cls._global_instance


class ReaderStore(PluginStore["ReaderPlugin"]):
    """Class that stores all the reader plugins."""

    def add_reader(self, reader: ReaderPlugin):
        self._plugin_items.append(reader)

    def iter_readers(
        self, path: Path | list[Path], min_priority: int = 0
    ) -> Iterator[tuple[str, ReaderPlugin]]:
        for reader in self._plugin_items:
            if reader.priority < min_priority:
                continue
            try:
                out = reader.match_model_type(path)
            except Exception as e:
                _warn_failed_provider(reader, e)
            else:
                if out is None:
                    _LOGGER.debug("Reader %r did not match", reader)
                else:
                    yield out, reader

    def get(
        self,
        path: Path | list[Path],
        empty_ok: bool = False,
        min_priority: int = 0,
    ) -> list[ReaderPlugin]:
        """List of reader plugins that can read the path."""
        matched = self._get_impl(_remove_tilde(path), min_priority=min_priority)
        _LOGGER.debug("Matched readers: %r", matched)
        if not matched and not empty_ok:
            if isinstance(path, list):
                msg = [p.name for p in path]
            else:
                msg = path.name
            raise ValueError(f"No reader functions available for {msg!r}")
        return matched

    def pick(
        self,
        path: Path,
        *,
        plugin: str | None = None,
        min_priority: int = 0,
    ) -> ReaderPlugin:
        """Pick a reader that match the inputs."""
        if plugin is not None:
            # if plugin is given, force to use it
            min_priority = -float("inf")
        return _pick_from_list(self.get(path, min_priority=min_priority), plugin)

    def run(
        self,
        path: Path | list[Path],
        *,
        plugin: str | None = None,
        min_priority: int = -float("inf"),
    ) -> WidgetDataModel:
        reader = self.pick(path, plugin=plugin, min_priority=min_priority)
        return reader.read(path)

    def _get_impl(
        self,
        path: Path | list[Path],
        min_priority: int = 0,
    ) -> list[ReaderPlugin]:
        matched: list[ReaderPlugin] = []
        for reader in self._plugin_items:
            if reader.priority < min_priority:
                continue
            try:
                out = reader.match_model_type(path)
            except Exception as e:
                _warn_failed_provider(reader, e)
            else:
                if out is None:
                    _LOGGER.debug("%r did not match", reader)
                else:
                    matched.append(reader)
        return matched


class WriterStore(PluginStore["WriterPlugin"]):
    def add_writer(self, reader: WriterPlugin):
        self._plugin_items.append(reader)

    def get(
        self,
        model: WidgetDataModel,
        path: Path,
        empty_ok: bool = False,
        min_priority: int = 0,
    ) -> list[WriterPlugin]:
        """List of writer plugins that can write the model."""
        matched = self._get_impl(model, _remove_tilde(path), min_priority=min_priority)
        if not matched and not empty_ok:
            raise ValueError(f"No writer functions available for {model.type!r}")
        return matched

    def pick(
        self,
        model: WidgetDataModel,
        path: Path,
        plugin: str | None = None,
        min_priority: int = 0,
    ) -> WriterPlugin:
        """Pick a writer that match the inputs to write the model."""
        if plugin is not None:
            # if plugin is given, force to use it
            min_priority = -float("inf")
        return _pick_from_list(self.get(model, path, min_priority=min_priority), plugin)

    def run(
        self,
        model: WidgetDataModel,
        path: Path,
        *,
        plugin: str | None = None,
        min_priority: int = -float("inf"),
    ) -> None:
        writer = self.pick(model, path, plugin=plugin, min_priority=min_priority)
        return writer.write(model, path)

    def _get_impl(
        self,
        model: WidgetDataModel,
        path: Path,
        min_priority: int = 0,
    ) -> list[WriterPlugin]:
        matched: list[WriterPlugin] = []
        for writer in self._plugin_items:
            if writer.priority < min_priority:
                continue
            try:
                out = writer.match_input(model, path)
            except Exception as e:
                _warn_failed_provider(writer, e)
            else:
                if not out:
                    _LOGGER.debug("%r did not match", writer)
                else:
                    matched.append(writer)
        return matched


def _pick_by_priority(tuples: list[_S]) -> _S:
    return max(tuples, key=lambda x: x.priority)


def _pick_from_list(choices: list[_S], plugin: str | None) -> _S:
    if plugin is None:
        out = _pick_by_priority(choices)
    else:
        plugin_info = PluginInfo.from_str(plugin)
        for each in choices:
            if each.plugin == plugin_info:
                out = each
                break
        else:
            warnings.warn(
                f"Plugin {plugin} not found, using the default one.",
                UserWarning,
                stacklevel=2,
            )
            out = _pick_by_priority(choices)
    _LOGGER.debug("Picked: %r", out)
    return out


def _warn_failed_provider(plugin_obj, e: Exception):
    return _LOGGER.error(f"Error in {plugin_obj!r}: {e}")


def _remove_tilde(path: Path | list[Path]) -> Path | list[Path]:
    try:
        if isinstance(path, list):
            return [p.with_name(p.name.rstrip("~")) for p in path]
        return path.with_name(path.name.rstrip("~"))
    except Exception:
        return path



from __future__ import annotations

from typing import TYPE_CHECKING, Literal
import numpy as np
from psygnal import Signal
from qtpy import QtCore, QtGui, QtWidgets as QtW

from himena.consts import DefaultFontFamily
from ._base import QBaseGraphicsScene, QBaseGraphicsView
from himena.qt._qlineedit import QDoubleLineEdit
from himena_builtins.qt.widgets._shared import quick_min_max

if TYPE_CHECKING:
    from numpy import ndarray as NDArray


class QHistogramView(QBaseGraphicsView):
    """Graphics view for displaying histograms and setting contrast limits."""

    clim_changed = QtCore.Signal(tuple)

    def __init__(self):
        super().__init__()
        self._hist_items = [self.addItem(QHistogramItem())]
        self._line_low = self.addItem(QClimLineItem(0))
        self._line_high = self.addItem(QClimLineItem(1))
        self._line_low.valueChanged.connect(self._on_clim_changed)
        self._line_high.valueChanged.connect(self._on_clim_changed)
        self._view_range: tuple[float, float] = (0.0, 1.0)
        self._minmax = (0.0, 1.0)  # limits of the movable range
        self._pos_drag_start: QtCore.QPoint | None = None
        self._default_hist_scale = "linear"

    def _on_clim_changed(self):
        clim = self.clim()
        if self._view_range is not None:
            v0, v1 = self._view_range
            x0, x1 = self.clim()
            if x0 < v0 or x1 > v1:
                self._view_range = clim
                self.update()
        self.clim_changed.emit(clim)

    def clim(self) -> tuple[float, float]:
        return tuple(sorted([self._line_low.value(), self._line_high.value()]))

    def set_clim(self, clim: tuple[float, float]):
        self._line_low.setValue(max(clim[0], self._minmax[0]))
        self._line_high.setValue(min(clim[1], self._minmax[1]))

    def set_minmax(self, minmax: tuple[float, float]):
        self._minmax = minmax
        self._line_low.setRange(*minmax)
        self._line_high.setRange(*minmax)

    def set_hist_for_array(
        self,
        arr: NDArray[np.number],
        clim: tuple[float, float],
        is_rgb: bool = False,
        color: QtGui.QColor = QtGui.QColor(100, 100, 100),
    ):
        # coerce the number of histogram items
        n_hist = 3 if is_rgb else 1
        for _ in range(n_hist, len(self._hist_items)):
            self.scene().removeItem(self._hist_items[-1])
            self._hist_items.pop()
        for _ in range(len(self._hist_items), n_hist):
            hist_item = QHistogramItem().with_hist_scale_func(self._default_hist_scale)
            self._hist_items.append(self.addItem(hist_item))

        minmax_fallback = clim
        if is_rgb:
            brushes = [
                QtGui.QBrush(QtGui.QColor(255, 0, 0, 128)),
                QtGui.QBrush(QtGui.QColor(0, 255, 0, 128)),
                QtGui.QBrush(QtGui.QColor(0, 0, 255, 255)),
            ]  # RGB
            for i, (item, brush) in enumerate(zip(self._hist_items, brushes)):
                item.with_brush(brush)
                item.set_hist_for_array(arr[:, :, i])
        else:
            brushes = [QtGui.QBrush(color)]
            self._hist_items[0].with_brush(brushes[0])
            # this fallback is needed when slider is changed.
            minmax_fallback = self._hist_items[0].set_hist_for_array(arr)

        if arr.dtype.kind in "ui":
            self.set_minmax((np.iinfo(arr.dtype).min, np.iinfo(arr.dtype).max))
        elif arr.dtype.kind == "b":
            self.set_minmax((0, 1))
        else:
            self.set_minmax(minmax_fallback)
        self.set_clim(clim)
        if self._view_range is None:
            self._view_range = clim

    def setValueFormat(self, fmt: str):
        self._line_low.setValueFormat(fmt)
        self._line_high.setValueFormat(fmt)

    def viewRect(self, width: float | None = None) -> QtCore.QRectF:
        """The current view range as a QRectF."""
        x0, x1 = self._view_range
        if width is None:
            width = x1 - x0
        return QtCore.QRectF(x0 - width * 0.03, 0, width * 1.06, 1)

    def set_view_range(self, x0: float, x1: float):
        self._view_range = (x0, x1)
        self.fitInView(self.viewRect(), QtCore.Qt.AspectRatioMode.IgnoreAspectRatio)

    def resizeEvent(self, event: QtGui.QResizeEvent):
        super().resizeEvent(event)
        self.fitInView(self.viewRect(), QtCore.Qt.AspectRatioMode.IgnoreAspectRatio)

    def showEvent(self, event: QtGui.QShowEvent):
        super().showEvent(event)
        x0, x1 = self.clim()
        self.fitInView(
            self.viewRect(x1 - x0), QtCore.Qt.AspectRatioMode.IgnoreAspectRatio
        )
        self._line_low.setValue(self._line_low.value())
        self._line_high.setValue(self._line_high.value())

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent):
        self._reset_view()

    def _reset_view(self):
        rect = self._line_low.boundingRect().united(self._line_high.boundingRect())
        for hist in self._hist_items:
            rect = rect.united(hist.boundingRect())
        x0, x1 = rect.left(), rect.right()
        self.fitInView(
            self.viewRect(x1 - x0), QtCore.Qt.AspectRatioMode.IgnoreAspectRatio
        )
        self._view_range = (x0, x1)

    def wheelEvent(self, event: QtGui.QWheelEvent):
        delta = event.angleDelta().y()
        if delta > 0:
            factor = 1.1
        else:
            factor = 1 / 1.1
        x0, x1 = self._view_range
        xcursor = self.mapToScene(event.pos()).x()
        x0 = max((x0 - xcursor) / factor + xcursor, self._minmax[0])
        x1 = min((x1 - xcursor) / factor + xcursor, self._minmax[1])
        self.set_view_range(x0, x1)

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        self._pos_drag_start = event.pos()
        self._pos_drag_prev = self._pos_drag_start
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        if self.scene().grabSource():
            return super().mouseMoveEvent(event)
        if event.buttons() & QtCore.Qt.MouseButton.LeftButton:
            pos = event.pos()
            if self._pos_drag_prev is not None:
                delta = self.mapToScene(pos) - self.mapToScene(self._pos_drag_prev)
                delta = delta.x()
                x0, x1 = self._view_range
                x0 -= delta
                x1 -= delta
                self.set_view_range(x0, x1)
            self._pos_drag_prev = pos
        return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        if (
            event.button() == QtCore.Qt.MouseButton.RightButton
            and self._pos_drag_start is not None
            and (self._pos_drag_start - event.pos()).manhattanLength() < 8
        ):
            # right clicked
            menu = self._make_context_menu()
            menu.exec(self.mapToGlobal(event.pos()))

        self._pos_drag_start = None
        self._pos_drag_prev = None
        self.scene().setGrabSource(None)
        return super().mouseReleaseEvent(event)

    def _make_context_menu(self) -> QtW.QMenu:
        menu = QtW.QMenu(self)
        menu.addAction("Reset view", self._reset_view)
        menu_scale = menu.addMenu("Scale")
        for scale in ["linear", "log"]:
            ac = menu_scale.addAction(
                scale.title(), lambda s=scale: self._set_hist_scale_func(s)
            )
            ac.setCheckable(True)
            ac.setChecked(scale == self._default_hist_scale)
        menu.addSeparator()
        menu.addAction("Copy Histogram", self._img_to_clipboard)
        return menu

    def _set_hist_scale_func(self, scale: Literal["linear", "log"]):
        for hist in self._hist_items:
            hist.with_hist_scale_func(scale)
        self._default_hist_scale = scale

    def _img_to_clipboard(self):
        img = self.grab().toImage()
        QtW.QApplication.clipboard().setImage(img)


class QClimLineItem(QtW.QGraphicsRectItem):
    """The line item for one of the contrast limits."""

    # NOTE: To properly set the bounding rect, we need to inherit from QGraphicsRectItem
    # with updated boundingRect method.
    valueChanged = Signal(float)
    _Y_LOW = -1
    _Y_HIGH = 2
    _WIDTH_NORMAL = 2
    _WIDTH_HOVER = 4

    def __init__(self, x: float):
        super().__init__()
        self._color = QtGui.QColor(255, 0, 0, 150)
        pen = QtGui.QPen(self._color, self._WIDTH_NORMAL)
        pen.setCosmetic(True)
        self._qpen = pen
        self.setZValue(1000)
        self.setFlag(QtW.QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self._is_dragging = False
        self._range = (-float("inf"), float("inf"))
        self._value = x
        self._value_fmt = ".1f"
        self.setCursor(QtCore.Qt.CursorShape.SizeHorCursor)
        self._value_label = QtW.QGraphicsSimpleTextItem()
        self._value_label.setFlag(
            QtW.QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations
        )
        self._value_label.setFont(QtGui.QFont(DefaultFontFamily, 8))
        self.setAcceptHoverEvents(True)

    def mousePressEvent(self, event: QtW.QGraphicsSceneMouseEvent):
        if event.button() == QtCore.Qt.MouseButton.LeftButton:
            self._is_dragging = True
            self.scene().setGrabSource(self)
        elif event.buttons() & QtCore.Qt.MouseButton.RightButton:
            self.scene().setGrabSource(self)
            menu = QClimMenu(self.scene().views()[0], self)
            menu._edit.setFocus()
            menu.exec(event.screenPos())
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QtW.QGraphicsSceneMouseEvent):
        if event.buttons() & QtCore.Qt.MouseButton.LeftButton:
            if self._is_dragging:
                self._drag_event(event)

    def mouseReleaseEvent(self, event: QtW.QGraphicsSceneMouseEvent):
        self._is_dragging = False
        self.scene().setGrabSource(None)
        self.setValue(event.pos().x())
        return super().mouseReleaseEvent(event)

    def hoverEnterEvent(self, event):
        self._qpen.setWidthF(self._WIDTH_HOVER)
        self._show_value_label()
        self.update()
        return super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        self._qpen.setWidthF(self._WIDTH_NORMAL)
        self._value_label.hide()
        self.update()
        return super().hoverLeaveEvent(event)

    def _show_value_label(self):
        txt = format(self.value(), self._value_fmt)
        self._value_label.setText(txt)
        vp = self.scene().views()[0].viewport()
        background_color = vp.palette().color(vp.backgroundRole())

        brightness = (
            0.299 * background_color.red()
            + 0.587 * background_color.green()
            + 0.114 * background_color.blue()
        )
        if brightness > 127:
            self._value_label.setBrush(QtGui.QBrush(QtCore.Qt.GlobalColor.black))
        else:
            self._value_label.setBrush(QtGui.QBrush(QtCore.Qt.GlobalColor.white))
        text_width = self._value_label.boundingRect().width()
        pos = QtCore.QPointF(self.value(), 0)
        if pos.x() + text_width / self._x_scale() > self._range[1]:
            pos.setX(pos.x() - (text_width + 4) / self._x_scale())
        else:
            pos.setX(pos.x() + 4 / self._x_scale())
        self._value_label.setPos(self.mapToScene(pos))
        if self._value_label.scene() is None:
            # prevent scene movement during adding the label
            rect = self.scene().sceneRect()
            self.scene().addItem(self._value_label)
            self.scene().setSceneRect(rect)
        self._value_label.show()

    def _drag_event(self, event: QtW.QGraphicsSceneMouseEvent):
        self.setValue(event.pos().x())
        self._show_value_label()
        if scene := self.scene():
            scene.update()

    def setValueFormat(self, fmt: str):
        self._value_fmt = fmt

    def paint(self, painter, option, widget):
        painter.setPen(self._qpen)
        start = QtCore.QPointF(self._value, self._Y_LOW)
        end = QtCore.QPointF(self._value, self._Y_HIGH)
        line = QtCore.QLineF(start, end)
        painter.drawLine(line)

    def value(self) -> float:
        """The x value of the line (the contrast limit)."""
        return self._value

    def setValue(self, x: float):
        """Set the x value of the line (the contrast limit)."""
        old_bbox = self.boundingRect()
        old_value = self._value
        new_value = min(max(x, self._range[0]), self._range[1])
        self._value = new_value
        new_bbox = self.boundingRect()
        self.setRect(new_bbox)
        if new_value != old_value:
            self.valueChanged.emit(self._value)
        if scene := self.scene():
            scene.update(self.mapRectToScene(old_bbox.united(new_bbox)))

    def setRange(self, low: float, high: float):
        """Set the min/max range of the line x value."""
        self._range = (low, high)
        if not low <= self.value() <= high:
            self.setValue(self.value())

    def scene(self) -> QBaseGraphicsScene:
        return super().scene()

    def _x_scale(self) -> float:
        return self.view().transform().m11()

    def view(self) -> QHistogramView:
        return self.scene().views()[0]

    def boundingRect(self):
        w = 10.0 / self._x_scale()
        x = self.value()
        return QtCore.QRectF(x - w / 2, self._Y_LOW, w, self._Y_HIGH - self._Y_LOW)


class QHistogramItem(QtW.QGraphicsPathItem):
    def __init__(self):
        super().__init__()
        self._hist_path = QtGui.QPainterPath()
        self._hist_brush = QtGui.QBrush(QtGui.QColor(100, 100, 100))
        self.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0, 0)))
        self._edges = np.array([0.0, 1.0])
        self._hist_values = np.zeros(0)
        self._hist_scale_func = _linear_scale

    def with_brush(self, brush: QtGui.QBrush) -> QHistogramItem:
        self._hist_brush = brush
        return self

    def with_hist_scale_func(self, scale: Literal["linear", "log"]) -> QHistogramItem:
        if scale == "linear":
            self._hist_scale_func = _linear_scale
        elif scale == "log":
            self._hist_scale_func = _log_scale
        else:
            raise ValueError(f"Unknown histogram scale: {scale}")
        self._update_histogram_path()
        return self

    def set_hist_for_array(self, arr: NDArray[np.number]) -> tuple[float, float]:
        _min, _max = quick_min_max(arr)
        if arr.dtype in ("int8", "uint8"):
            _nbin = 64
        else:
            _nbin = 256
        # nbin should not be more than half of the number of pixels
        _nbin = min(_nbin, int(np.prod(arr.shape[-2:])) // 2)
        # draw histogram
        if arr.dtype.kind == "b":
            edges = np.array([0, 0.5, 1])
            frac_true = np.sum(arr) / arr.size
            hist = np.array([1 - frac_true, frac_true])
        elif _max > _min:
            arr = arr.clip(_min, _max)
            if arr.dtype.kind in "ui":
                _nbin = int(_max - _min) // max(int(np.ceil((_max - _min) / _nbin)), 1)
            normed = ((arr - _min) / (_max - _min) * _nbin).astype(np.uint8)
            hist = np.bincount(normed.ravel(), minlength=_nbin)
            edges = np.linspace(_min, _max, _nbin + 1)
        else:
            edges = np.array([_min, _max])
            hist = np.zeros(1)
        self._edges = edges
        self._hist_values = hist

        self._update_histogram_path()
        return _min, _max

    def _update_histogram_path(self):
        _path = QtGui.QPainterPath()
        edges = self._edges
        hist = self._hist_values
        self.setBrush(self._hist_brush)
        _path.moveTo(edges[0], 1)
        for e0, e1, h in zip(edges[:-1], edges[1:], self._hist_scale_func(hist)):
            _path.lineTo(e0, 1 - h)
            _path.lineTo(e1, 1 - h)
        _path.lineTo(edges[-1], 1)
        _path.closeSubpath()
        self.setPath(_path)
        self.update()


def _linear_scale(hist: NDArray[np.number]) -> NDArray[np.number]:
    if hist.size > 0 and (hmax := hist.max()) > 0:
        return hist / hmax
    return hist


def _log_scale(hist: NDArray[np.number]) -> NDArray[np.number]:
    hist_log = np.log(hist + 1)
    if hist_log.size > 0 and (hmax := hist_log.max()) > 0:
        return hist_log / hmax
    return hist_log


class QClimMenu(QtW.QMenu):
    def __init__(self, parent: QHistogramView, item: QClimLineItem):
        super().__init__(parent)
        self._hist_view = parent
        self._item = item
        self._edit = QDoubleLineEdit()
        self._edit.setText(format(item.value(), item._value_fmt))
        self._edit.editingFinished.connect(self._on_value_changed)
        widget_action = QtW.QWidgetAction(self)
        widget_action.setDefaultWidget(self._edit)
        self.addAction(widget_action)

    def _on_value_changed(self):
        value = float(self._edit.text())
        self._item.setValue(value)
        # update min/max
        if value < self._hist_view._minmax[0]:
            self._hist_view.set_minmax((value, self._hist_view._minmax[1]))
        elif value > self._hist_view._minmax[1]:
            self._hist_view.set_minmax((self._hist_view._minmax[0], value))
        # update view range
        v0, v1 = self._hist_view._view_range
        if value < v0:
            self._hist_view.set_view_range(value, v1)
        elif value > v1:
            self._hist_view.set_view_range(v0, value)
        self.close()



"""The standardized, inter-changable way to represent specific data in Himena."""

from himena.standards._base import BaseMetadata, read_metadata, write_metadata

__all__ = ["BaseMetadata", "read_metadata", "write_metadata"]
